from datetime import timedelta
from os import PathLike
from pathlib import Path

from geojson import Point
import numpy
from shapely.geometry import LineString

from packetraven.packets import APRSPacket
from packetraven.packets.tracks import APRSTrack, LocationPacketTrack

KML_STANDARD = '{http://www.opengis.net/kml/2.2}'


def write_packet_tracks(packet_tracks: [LocationPacketTrack], output_filename: PathLike):
    if not isinstance(output_filename, Path):
        output_filename = Path(output_filename)
    output_filename = output_filename.resolve().expanduser()
    if output_filename.suffix == '.txt':
        packets = []
        for packet_track_index, packet_track in enumerate(packet_tracks):
            packets.extend(packet_track)
        packets = sorted(packets)
        lines = [
            f'{packet.time:%Y-%m-%d %H:%M:%S %Z}: {packet.frame}'
            if isinstance(packet, APRSPacket)
            else f'{packet.time:%Y-%m-%d %H:%M:%S %Z}'
            for packet in packets
        ]
        with open(output_filename, 'w') as output_file:
            output_file.write('\n'.join(lines))
    elif output_filename.suffix == '.geojson':
        import geojson

        features = []
        for packet_track in packet_tracks:
            ascent_rates = numpy.round(packet_track.ascent_rates, 3)
            ground_speeds = numpy.round(packet_track.ground_speeds, 3)

            for packet_index, packet in enumerate(packet_track):
                properties = {
                    'time': f'{packet.time:%Y%m%d%H%M%S}',
                    'altitude': packet.coordinates[2],
                    'ascent_rate': ascent_rates[packet_index],
                    'ground_speed': ground_speeds[packet_index],
                }

                properties.update(packet.attributes)

                features.append(
                    geojson.Feature(
                        geometry=geojson.Point(packet.coordinates.tolist()),
                        properties=properties,
                    )
                )

            properties = {
                'time': f'{packet_track.packets[-1].time:%Y%m%d%H%M%S}',
                'altitude': packet_track.coordinates[-1, -1],
                'ascent_rate': ascent_rates[-1],
                'ground_speed': ground_speeds[-1],
                'seconds_to_ground': packet_track.time_to_ground / timedelta(seconds=1),
            }

            if isinstance(packet_track, APRSTrack):
                properties['callsign'] = packet_track.callsign

            features.append(
                geojson.Feature(
                    geometry=geojson.LineString(
                        [packet.coordinates.tolist() for packet in packet_track.packets]
                    ),
                    properties=properties,
                )
            )

        features = geojson.FeatureCollection(features)

        with open(output_filename, 'w') as output_file:
            geojson.dump(features, output_file)
    elif output_filename.suffix == '.kml':
        from fastkml import kml

        output_kml = kml.KML()
        document = kml.Document(
            KML_STANDARD, '1', 'root document', 'root document, containing geometries'
        )
        output_kml.append(document)

        for packet_track_index, packet_track in enumerate(packet_tracks):
            ascent_rates = numpy.round(packet_track.ascent_rates, 3)
            ground_speeds = numpy.round(packet_track.ground_speeds, 3)

            for packet_index, packet in enumerate(packet_track):
                placemark = kml.Placemark(
                    KML_STANDARD,
                    f'1 {packet_track_index} {packet_index}',
                    f'{packet.time:%Y%m%d%H%M%S} {packet_track.callsign if isinstance(packet_track, APRSTrack) else ""}',
                    f'altitude={packet.coordinates[2]} '
                    f'ascent_rate={ascent_rates[packet_index]} '
                    f'ground_speed={ground_speeds[packet_index]}',
                )
                placemark.geometry = Point(packet.coordinates.tolist())
                document.append(placemark)

            placemark = kml.Placemark(
                KML_STANDARD,
                f'1 {packet_track_index}',
                packet_track.callsign if isinstance(packet_track, APRSTrack) else '',
                f'altitude={packet_track.coordinates[-1, -1]} '
                f'ascent_rate={ascent_rates[-1]} '
                f'ground_speed={ground_speeds[-1]} '
                f'seconds_to_ground={packet_track.time_to_ground / timedelta(seconds=1)}',
            )
            placemark.geometry = LineString(packet_track.coordinates)
            document.append(placemark)

        with open(output_filename, 'w') as output_file:
            output_file.write(output_kml.to_string())
    else:
        raise NotImplementedError(
            f'saving to file type "{output_filename.suffix}" has not been implemented'
        )



from datetime import datetime

from serial import Serial

from packetraven.connections.base import (
    APRSPacketSource,
    LOGGER,
    next_open_serial_port,
    TimeIntervalError,
)
from packetraven.packets import APRSPacket


class SerialTNC(APRSPacketSource):
    def __init__(self, serial_port: str = None, callsigns: [str] = None):
        """
        Connect to TNC over given serial port.

        :param serial_port: port name
        :param callsigns: list of callsigns to return from source
        """

        if serial_port is None or serial_port == '' or serial_port == 'auto':
            try:
                serial_port = next_open_serial_port()
            except ConnectionError:
                raise ConnectionError('could not find TNC connected to serial')
        else:
            serial_port = serial_port.strip('"')

        self.serial_connection = Serial(serial_port, baudrate=9600, timeout=1)
        super().__init__(self.serial_connection.port, callsigns)
        self.__last_access_time = None

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )
        packets = []
        for line in self.serial_connection.readlines():
            try:
                packet = APRSPacket.from_frame(line, source=self.location)
                packets.append(packet)
            except Exception as error:
                LOGGER.error(f'{error.__class__.__name__} - {error}')
        if self.callsigns is not None:
            packets = [packet for packet in packets if packet.from_callsign in self.callsigns]
        self.__last_access_time = datetime.now()
        return packets

    def close(self):
        self.serial_connection.close()

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.location)}, {repr(self.callsigns)})'



from datetime import datetime, timedelta
from typing import Any, Union

from dateutil.parser import parse as parse_date
import numpy
from pyproj import CRS, Geod, Transformer

from packetraven.packets.parsing import InvalidPacketError, parse_raw_aprs

DEFAULT_CRS = CRS.from_epsg(4326)


class LocationPacket:
    """ location packet encoding (x, y, z) and time """

    def __init__(
        self,
        time: datetime,
        x: float,
        y: float,
        z: float = None,
        crs: CRS = None,
        source: str = None,
        **kwargs,
    ):
        if isinstance(time, str):
            time = parse_date(time)

        self.time = time
        self.coordinates = numpy.array((x, y, z if z is not None else 0))
        self.crs = crs if crs is not None else DEFAULT_CRS
        self.source = source
        self.attributes = kwargs

    def overground_distance(self, point: (float, float)) -> float:
        """
        horizontal distance over ellipsoid

        :param point: (x, y) point
        :return: distance in ellipsodal units
        """

        if not isinstance(point, numpy.ndarray):
            point = numpy.array(point)
        coordinates = numpy.stack([self.coordinates[:2], point], axis=0)
        if self.crs.is_projected:
            return numpy.hypot(*numpy.sum(numpy.diff(coordinates, axis=0), axis=0))
        else:
            ellipsoid = self.crs.datum.to_json_dict()['ellipsoid']
            geodetic = Geod(a=ellipsoid['semi_major_axis'], rf=ellipsoid['inverse_flattening'])
            return geodetic.line_length(coordinates[:, 0], coordinates[:, 1])

    def transform_to(self, crs: CRS):
        transformer = Transformer.from_crs(self.crs, crs)
        self.coordinates = transformer.transform(self.coordinates)

    def __getitem__(self, field: str) -> Any:
        if field not in self:
            raise KeyError(f'"{field}" not in packet')
        return {**self.__dict__, **self.attributes}[field]

    def __setitem__(self, field: str, value: Any):
        self.attributes[field] = value

    def __contains__(self, field: str) -> bool:
        return field in {**self.__dict__, **self.attributes}

    def __sub__(self, other: 'LocationPacket') -> 'Distance':
        return Distance.from_packets(self, other)

    def __eq__(self, other: 'LocationPacket') -> bool:
        return numpy.allclose(self.coordinates, other.coordinates)

    def __gt__(self, other: 'LocationPacket') -> bool:
        return self.time > other.time

    def __lt__(self, other: 'LocationPacket') -> bool:
        return self.time < other.time

    def __str__(self) -> str:
        return f'{self.time} {self.coordinates}'

    def __repr__(self) -> str:
        coordinate_string = ', '.join(
            f'{key}={value}'
            for key, value in dict(zip(('x', 'y', 'z'), self.coordinates)).items()
        )
        attribute_string = ', '.join(
            f'{attribute}={repr(value)}' for attribute, value in self.attributes.items()
        )
        return (
            f'{self.__class__.__name__}(time={repr(self.time)}, {coordinate_string}, '
            f'crs={self.crs.__class__.__name__}.from_epsg({repr(self.crs.to_epsg())}), {attribute_string})'
        )


class Distance:
    def __init__(self, interval: timedelta, horizontal: float, vertical: float, crs: CRS):
        self.__interval = interval
        self.__horizontal = horizontal
        self.__vertical = vertical
        self.__crs = crs

    @classmethod
    def from_packets(cls, packet_1: LocationPacket, packet_2: LocationPacket):
        """
        Get distance between two location packets, using the first packet's CRS.

        :param packet_1: first location packet
        :param packet_2: second location packet
        """

        if packet_2.crs != packet_1.crs:
            transformer = Transformer.from_crs(packet_2.crs, packet_1.crs)
            packet_2_coordinates = transformer.transform(*packet_2.coordinates)
        else:
            packet_2_coordinates = packet_2.coordinates

        interval = packet_1.time - packet_2.time
        horizontal_distance = packet_1.overground_distance(packet_2_coordinates[:2])
        vertical_distance = packet_1.coordinates[2] - packet_2_coordinates[2]

        return cls(interval, horizontal_distance, vertical_distance, packet_1.crs)

    @property
    def interval(self) -> timedelta:
        return self.__interval

    @property
    def seconds(self) -> float:
        return self.interval / timedelta(seconds=1)

    @property
    def overground(self) -> float:
        return self.__horizontal

    @property
    def ascent(self) -> float:
        return self.__vertical

    @property
    def crs(self) -> CRS:
        return self.__crs

    @property
    def ascent_rate(self) -> float:
        return self.__vertical / self.seconds if self.seconds > 0 else 0

    @property
    def ground_speed(self) -> float:
        return self.__horizontal / self.seconds if self.seconds > 0 else 0

    def __str__(self) -> str:
        return f'{self.seconds}s, {self.ascent:6.2f}m vertical, {self.overground:6.2f}m horizontal'

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}({repr(self.interval)}, {self.ascent}, {self.overground}, '
            f'{self.crs.__class__.__name__}.from_epsg({repr(self.crs.to_epsg())}))'
        )


class APRSPacket(LocationPacket):
    """ APRS packet containing parsed APRS fields, along with location and time """

    def __init__(
        self,
        from_callsign: str,
        to_callsign: str,
        time: datetime,
        x: float,
        y: float,
        z: float = None,
        crs: CRS = None,
        **kwargs,
    ):
        """
        APRS packet object from raw packet and given datetime

        :param from_callsign: originating callsign of packet
        :param to_callsign: destination callsign of packet
        :param x: x value
        :param y: y value
        :param z: z value
        :param crs: coordinate reference system
        :param time: time of packet, either as datetime object, seconds since Unix epoch, or ISO format date string.
        """

        kwargs['from'] = from_callsign
        kwargs['to'] = to_callsign if to_callsign is not None else 'APRS'
        super().__init__(time, x, y, z, crs, **kwargs)

    @classmethod
    def from_frame(
        cls, frame: Union[str, bytes, dict], packet_time: datetime = None, **kwargs
    ) -> 'APRSPacket':
        """
        APRS packet object from raw packet and given datetime

        :param frame: string containing raw packet
        :param packet_time: Time of packet, either as datetime object, seconds since Unix epoch, or ISO format date string.
        """

        # parse packet with metric units
        parsed_packet = parse_raw_aprs(frame)

        if 'longitude' not in parsed_packet or 'latitude' not in parsed_packet:
            raise InvalidPacketError(f'Input packet does not contain location data: {frame}')

        if packet_time is None:
            if 'timestamp' in parsed_packet:
                # extract time from Unix epoch
                packet_time = datetime.fromtimestamp(float(parsed_packet['timestamp']))
            else:
                # TODO make HABduino add timestamp to packet upon transmission
                # otherwise default to time the packet was received (now)
                packet_time = datetime.now()

        return cls(
            parsed_packet['from'],
            parsed_packet['to'],
            packet_time,
            parsed_packet['longitude'],
            parsed_packet['latitude'],
            parsed_packet['altitude'] if 'altitude' in parsed_packet else None,
            crs=DEFAULT_CRS,
            **{
                key: value
                for key, value in parsed_packet.items()
                if key not in ['from', 'to', 'longitude', 'latitude', 'altitude']
            },
            **kwargs,
        )

    @property
    def from_callsign(self) -> str:
        return self['from']

    @from_callsign.setter
    def from_callsign(self, callsign: str):
        self['from'] = callsign

    @property
    def to_callsign(self) -> str:
        return self['to']

    @to_callsign.setter
    def to_callsign(self, callsign: str):
        self['to'] = callsign

    @property
    def frame(self) -> str:
        if 'raw' in self and self['raw'] is not None:
            frame = self['raw']
        else:
            # https://aprs-python.readthedocs.io/en/stable/parse_formats.html#normal
            x, y, z = self.coordinates
            north = y >= 0
            east = x >= 0
            x = abs(x)
            y = abs(y)
            frame = f'{self.from_callsign}>{self["to"]}'
            if 'path' in self:
                frame += f',{",".join(self["path"])}'
            if 'via' in self:
                frame += f',{self["via"]}'
            frame += f':!{y * 100:04.2f}{"N" if north else "S"}/{x * 100:05.2f}{"E" if east else "W"}-/A={z * 3.28084:06.0f}'
            if 'comment' in self:
                frame += f' {self["comment"]}'
        return frame

    def __getitem__(self, field: str) -> Any:
        if field == 'callsign':
            field = 'from'
        return super().__getitem__(field)

    def __setitem__(self, field: str, value: Any):
        if field == 'path' and isinstance(value, str):
            value = value.split(',')
        return super().__setitem__(field, value)

    def __contains__(self, field: str):
        if field == 'callsign':
            field = 'from'
        return super().__contains__(field)

    def __iter__(self):
        yield from self.attributes

    def __eq__(self, other: 'APRSPacket') -> bool:
        return (
            super().__eq__(other)
            and self.from_callsign == other.from_callsign
            and self['comment'] == other['comment']
        )

    def __str__(self) -> str:
        return self.frame

    def __repr__(self) -> str:
        coordinate_string = ', '.join(
            f'{key}={value}'
            for key, value in dict(zip(('x', 'y', 'z'), self.coordinates)).items()
        )
        attribute_string = ', '.join(
            f'{attribute}={repr(value)}'
            for attribute, value in self.attributes.items()
            if attribute not in ['from', 'to']
        )
        return (
            f'{self.__class__.__name__}(from_callsign={repr(self.from_callsign)}, to_callsign={repr(self.to_callsign)}, time={repr(self.time)}, '
            f'{coordinate_string}, crs={self.crs.__class__.__name__}.from_epsg({repr(self.crs.to_epsg())}), {attribute_string})'
        )



import asyncio
from datetime import datetime, timedelta
import logging
from os import PathLike
from pathlib import Path
import re
import sys
from typing import Callable, Collection

from dateutil.parser import parse
import numpy
import teek
from teek.extras import more_dialogs

from packetraven import APRSDatabaseTable, APRSfi, RawAPRSTextFile, SerialTNC
from packetraven.__main__ import DEFAULT_INTERVAL_SECONDS, LOGGER, retrieve_packets
from packetraven.connections.base import available_serial_ports, next_open_serial_port
from packetraven.connections.file import PacketGeoJSON
from packetraven.connections.internet import APRSis
from packetraven.gui.plotting import LivePlot
from packetraven.packets import APRSPacket, LocationPacket
from packetraven.packets.tracks import LocationPacketTrack, PredictedTrajectory
from packetraven.packets.writer import write_packet_tracks
from packetraven.predicts import get_predictions, PredictionError
from packetraven.utilities import get_logger


class PacketRavenGUI:
    def __init__(
        self,
        callsigns: [str] = None,
        start_date: datetime = None,
        end_date: datetime = None,
        log_filename: PathLike = None,
        output_filename: PathLike = None,
        prediction_filename: PathLike = None,
        interval_seconds: int = None,
        igate: bool = False,
        **kwargs,
    ):
        main_window = teek.Window('PacketRaven')
        self.__windows = {'main': main_window}

        self.interval_seconds = (
            interval_seconds if interval_seconds is not None else DEFAULT_INTERVAL_SECONDS
        )

        self.__configuration = {
            'aprs_fi': {'aprs_fi_key': None},
            'tnc': {'tnc': None},
            'database': {
                'database_hostname': None,
                'database_database': None,
                'database_table': None,
                'database_username': None,
                'database_password': None,
            },
            'ssh_tunnel': {'ssh_hostname': None, 'ssh_username': None, 'ssh_password': None},
            'prediction': {
                'prediction_start_time': None,
                'prediction_start_location': None,
                'prediction_ascent_rate': None,
                'prediction_burst_altitude': None,
                'prediction_sea_level_descent_rate': None,
                'prediction_float_altitude': None,
                'prediction_float_duration': None,
                'prediction_api_url': None,
            },
        }

        for section_name, section in self.__configuration.items():
            section.update({key: value for key, value in kwargs.items() if key in section})

        self.igate = igate

        self.database = None
        self.aprs_is = None
        self.__connections = []

        self.__timeout = None
        self.__running = False
        self.__toggles = {}
        self.__packet_tracks = {}

        self.__frames = {}
        self.__elements = {}

        self.__plots = {}

        configuration_frame = teek.Frame(main_window)
        configuration_frame.grid(row=len(main_window.grid_rows), column=0, pady=10)
        self.__frames['configuration'] = configuration_frame

        start_date_entry = self.__add_entry_box(
            configuration_frame, title='start_date', label='Start Date', width=22, sticky='w',
        )
        self.__add_entry_box(
            configuration_frame,
            title='end_date',
            label='End Date',
            width=22,
            row=start_date_entry.grid_info()['row'],
            column=start_date_entry.grid_info()['column'] + 1,
            sticky='e',
        )

        separator = teek.Separator(configuration_frame, orient='horizontal')
        separator.grid(
            row=len(configuration_frame.grid_rows),
            column=0,
            columnspan=len(configuration_frame.grid_columns) + 1,
            sticky='ew',
            pady=10,
        )

        self.__add_entry_box(
            configuration_frame,
            title='callsigns',
            label='Callsigns',
            width=63,
            columnspan=len(configuration_frame.grid_columns),
        )
        self.__file_selection_option = 'select file(s)...'
        self.__add_combo_box(
            configuration_frame,
            title='tnc',
            label='TNC',
            options=list(available_serial_ports()) + [self.__file_selection_option],
            option_select=self.__select_tnc,
            width=60,
            columnspan=len(configuration_frame.grid_columns),
            sticky='w',
        )

        separator = teek.Separator(configuration_frame, orient='horizontal')
        separator.grid(
            row=len(configuration_frame.grid_rows),
            column=0,
            columnspan=len(configuration_frame.grid_columns) + 1,
            sticky='ew',
            pady=10,
        )

        log_file_label = teek.Label(configuration_frame, text='Log')
        log_file_label.grid(row=len(configuration_frame.grid_rows), column=0, sticky='w')

        log_file_frame = teek.Frame(configuration_frame)
        log_file_frame.grid(
            row=log_file_label.grid_info()['row'],
            column=1,
            columnspan=len(configuration_frame.grid_columns) - 1,
        )

        self.__toggles['log_file'] = teek.BooleanVar()
        self.__toggles['log_file'].set(log_filename is not None)
        log_file_checkbox = teek.Checkbutton(
            log_file_frame, variable=self.__toggles['log_file'], command=self.__toggle_log_file
        )
        log_file_checkbox.grid(row=0, column=0, padx=10)

        self.__elements['log_file_box'] = self.__add_file_box(
            log_file_frame,
            row=0,
            column=1,
            title='log_file',
            file_select=self.__select_log_file,
            width=52,
            sticky='w',
        )

        output_file_label = teek.Label(configuration_frame, text='Output')
        output_file_label.grid(row=len(configuration_frame.grid_rows), column=0, sticky='w')

        output_file_frame = teek.Frame(configuration_frame)
        output_file_frame.grid(
            row=output_file_label.grid_info()['row'],
            column=1,
            columnspan=len(configuration_frame.grid_columns) - 1,
        )

        self.__toggles['output_file'] = teek.BooleanVar()
        self.__toggles['output_file'].set(output_filename is not None)
        output_file_checkbox = teek.Checkbutton(
            output_file_frame,
            variable=self.__toggles['output_file'],
            command=self.__toggle_output_file,
        )
        output_file_checkbox.grid(row=0, column=0, padx=10)

        self.__elements['output_file_box'] = self.__add_file_box(
            output_file_frame,
            row=0,
            column=1,
            title='output_file',
            file_select=self.__select_output_file,
            width=52,
            sticky='w',
        )

        separator = teek.Separator(configuration_frame, orient='horizontal')
        separator.grid(
            row=len(configuration_frame.grid_rows),
            column=0,
            columnspan=len(configuration_frame.grid_columns) + 1,
            sticky='ew',
            pady=10,
        )

        prediction_label = teek.Label(configuration_frame, text='Predict')
        prediction_label.grid(row=len(configuration_frame.grid_rows), column=0, sticky='w')

        prediction_frame = teek.Frame(configuration_frame)
        prediction_frame.grid(
            row=prediction_label.grid_info()['row'],
            column=1,
            columnspan=len(configuration_frame.grid_columns) - 1,
        )

        self.__toggles['prediction_file'] = teek.BooleanVar()
        self.__toggles['prediction_file'].set(prediction_filename is not None)
        prediction_checkbox = teek.Checkbutton(
            prediction_frame,
            variable=self.__toggles['prediction_file'],
            command=self.__toggle_prediction_file,
        )
        prediction_checkbox.grid(row=0, column=0, padx=10)

        self.__elements['prediction_file_box'] = self.__add_file_box(
            prediction_frame,
            row=0,
            column=1,
            title='prediction_file',
            file_select=self.__select_prediction_file,
            width=52,
            sticky='w',
        )

        separator = teek.Separator(configuration_frame, orient='horizontal')
        separator.grid(
            row=len(configuration_frame.grid_rows),
            column=0,
            columnspan=len(configuration_frame.grid_columns) + 1,
            sticky='ew',
            pady=10,
        )

        plot_label = teek.Label(configuration_frame, text='Plots')
        plot_label.grid(row=len(configuration_frame.grid_rows), column=0, sticky='w')

        plot_checkbox_frame = teek.Frame(configuration_frame)
        plot_checkbox_frame.grid(
            row=plot_label.grid_info()['row'],
            column=0,
            columnspan=len(configuration_frame.grid_columns) - 1,
        )

        plot_variables = ['altitude', 'ascent_rate', 'ground_speed']
        self.__plot_toggles = {}
        for plot_index, plot in enumerate(plot_variables):
            boolean_var = teek.BooleanVar()
            boolean_var.set(False)
            plot_checkbox = teek.Checkbutton(
                plot_checkbox_frame, text=plot, variable=boolean_var
            )
            plot_checkbox.grid(row=0, column=plot_index, padx=10)
            self.__plot_toggles[plot] = boolean_var

        separator = teek.Separator(main_window, orient='horizontal')
        separator.grid(row=len(main_window.grid_rows), column=0, sticky='ew')

        control_frame = teek.Frame(main_window)
        control_frame.grid(row=len(main_window.grid_rows), column=0, pady=10)
        self.__frames['controls'] = control_frame

        self.__toggle_text = teek.StringVar()
        self.__toggle_text.set('Start')
        toggle_button = teek.Button(
            control_frame, textvariable=self.__toggle_text, command=self.toggle
        )
        toggle_button.grid(row=len(control_frame.grid_rows), column=0, sticky='nsew')
        self.__elements['toggle_button'] = toggle_button

        self.callsigns = callsigns
        self.tncs = self.__configuration['tnc']['tnc']

        if start_date is not None:
            self.start_date = start_date
        if end_date is not None:
            self.end_date = end_date

        self.log_filename = log_filename
        if self.log_filename is None:
            self.log_filename = Path('~') / 'Desktop'
        self.__toggle_log_file()

        self.output_filename = output_filename
        if self.output_filename is None:
            self.output_filename = Path('~') / 'Desktop'
        self.__toggle_output_file()

        self.prediction_filename = prediction_filename
        if self.prediction_filename is None:
            self.prediction_filename = Path('~') / 'Desktop'
        self.__toggle_prediction_file()
        self.__predictions = {}

        self.__windows['main'].on_delete_window.connect(self.close)

        teek.run()

    @property
    def callsigns(self) -> [str]:
        callsigns = self.__elements['callsigns'].text
        if len(callsigns) > 0:
            callsigns = [
                callsign.strip().upper()
                for callsign in re.split(',+\ *|\ +', callsigns.strip('"'))
            ]
        else:
            callsigns = None
        self.callsigns = callsigns
        return callsigns

    @callsigns.setter
    def callsigns(self, callsigns: [str]):
        if callsigns is not None:
            callsigns = ', '.join([callsign.upper() for callsign in callsigns])
        else:
            callsigns = ''
        self.__replace_text(self.__elements['callsigns'], callsigns)

    @property
    def tncs(self) -> [str]:
        """ locations of TNCs parsing APRS audio into ASCII frames """
        tncs = []
        for tnc in self.__elements['tnc'].text.split(','):
            tnc = tnc.strip()
            if len(tnc) > 0:
                if tnc.upper() == 'AUTO':
                    try:
                        tnc = next_open_serial_port()
                    except OSError:
                        LOGGER.warning(f'no open serial ports')
                        tnc = None
                tncs.append(tnc)
        return tncs

    @tncs.setter
    def tncs(self, filenames: [PathLike]):
        if filenames is None:
            filenames = []
        elif not isinstance(filenames, Collection) or isinstance(filenames, str):
            filenames = [filenames]

        filenames = [str(filename) for filename in filenames]

        self.__configuration['tnc']['tnc'] = filenames
        self.__replace_text(self.__elements['tnc'], ', '.join(filenames))

    @property
    def start_date(self) -> datetime:
        start_date = self.__elements['start_date'].text
        if len(start_date) > 0:
            start_date = parse(start_date)
        else:
            start_date = None
        return start_date

    @start_date.setter
    def start_date(self, start_date: datetime):
        if start_date is not None:
            if isinstance(start_date, str):
                start_date = parse(start_date)
            start_date = f'{start_date:%Y-%m-%d %H:%M:%S}'
        else:
            start_date = ''
        self.__replace_text(self.__elements['start_date'], start_date)

    @property
    def end_date(self) -> datetime:
        end_date = self.__elements['end_date'].text
        if len(end_date) > 0:
            end_date = parse(end_date)
        else:
            end_date = None
        return end_date

    @end_date.setter
    def end_date(self, end_date: datetime):
        if end_date is not None:
            if isinstance(end_date, str):
                end_date = parse(end_date)
            end_date = f'{end_date:%Y-%m-%d %H:%M:%S}'
        else:
            end_date = ''
        self.__replace_text(self.__elements['end_date'], end_date)

    @property
    def log_filename(self) -> Path:
        if self.toggles['log_file']:
            filename = self.__elements['log_file'].text
            if len(filename) > 0:
                filename = Path(filename)
                if filename.expanduser().resolve().is_dir():
                    self.log_filename = filename
                    filename = self.log_filename
            else:
                filename = None
        else:
            filename = None
        return filename

    @log_filename.setter
    def log_filename(self, filename: PathLike):
        if filename is not None:
            if not isinstance(filename, Path):
                filename = Path(filename)
            if filename.expanduser().resolve().is_dir():
                filename = filename / f'packetraven_log_{datetime.now():%Y%m%dT%H%M%S}.txt'
        else:
            filename = ''
        self.__replace_text(self.__elements['log_file'], filename)

    @property
    def output_filename(self) -> Path:
        if self.toggles['output_file']:
            filename = self.__elements['output_file'].text
            if len(filename) > 0:
                filename = Path(filename)
                if filename.expanduser().resolve().is_dir():
                    self.output_filename = filename
                    filename = self.output_filename
            else:
                filename = None
        else:
            filename = None
        return filename

    @output_filename.setter
    def output_filename(self, filename: PathLike):
        if filename is not None:
            if not isinstance(filename, Path):
                filename = Path(filename)
            if filename.expanduser().resolve().is_dir():
                filename = (
                    filename / f'packetraven_output_{datetime.now():%Y%m%dT%H%M%S}.geojson'
                )
        else:
            filename = ''
        self.__replace_text(self.__elements['output_file'], filename)

    @property
    def prediction_filename(self) -> Path:
        if self.toggles['prediction_file']:
            filename = self.__elements['prediction_file'].text
            if len(filename) > 0:
                filename = Path(filename)
                if filename.expanduser().resolve().is_dir():
                    self.prediction_filename = filename
                    filename = self.prediction_filename
            else:
                filename = None
        else:
            filename = None
        return filename

    @prediction_filename.setter
    def prediction_filename(self, filename: PathLike):
        if filename is not None:
            if not isinstance(filename, Path):
                filename = Path(filename)
            if filename.expanduser().resolve().is_dir():
                filename = (
                    filename / f'packetraven_predict_{datetime.now():%Y%m%dT%H%M%S}.geojson'
                )
        else:
            filename = ''
        self.__replace_text(self.__elements['prediction_file'], filename)

    @property
    def toggles(self) -> {str: bool}:
        return {key: value.get() for key, value in self.__toggles.items()}

    @property
    def running(self) -> bool:
        return self.__running

    @running.setter
    def running(self, running: bool):
        if running is not self.running:
            self.toggle()

    @property
    def packet_tracks(self) -> {str: LocationPacketTrack}:
        return self.__packet_tracks

    @property
    def predictions(self) -> {str: PredictedTrajectory}:
        return self.__predictions if self.toggles['prediction_file'] else None

    def toggle(self):
        if not self.running:
            self.__elements['toggle_button'].busy_hold()
            self.__toggle_text.set('Stop')

            if self.log_filename is not None:
                get_logger(LOGGER.name, self.log_filename)

            if self.toggles['log_file']:
                set_child_states(self.__elements['log_file_box'], 'disabled')

            start_date = self.start_date
            self.__elements['start_date'].config['state'] = 'disabled'

            end_date = self.end_date
            self.__elements['end_date'].config['state'] = 'disabled'

            callsigns = self.callsigns
            self.__elements['callsigns'].config['state'] = 'disabled'

            filter_message = 'retrieving packets'
            if start_date is not None and end_date is None:
                filter_message += f' sent after {start_date:%Y-%m-%d %H:%M:%S}'
            elif start_date is None and end_date is not None:
                filter_message += f' sent before {end_date:%Y-%m-%d %H:%M:%S}'
            elif start_date is not None and end_date is not None:
                filter_message += f' sent between {start_date:%Y-%m-%d %H:%M:%S} and {end_date:%Y-%m-%d %H:%M:%S}'
            if callsigns is not None:
                filter_message += f' from {len(callsigns)} callsigns: {callsigns}'
            LOGGER.info(filter_message)

            connection_errors = []
            try:
                tncs = self.tncs
                self.__elements['tnc'].config['state'] = 'disabled'
                for tnc in tncs:
                    try:
                        if Path(tnc).suffix in ['.txt', '.log']:
                            tnc = RawAPRSTextFile(tnc, self.callsigns)
                            LOGGER.info(f'reading file {tnc.location}')
                        else:
                            tnc = SerialTNC(tnc, self.callsigns)
                            LOGGER.info(f'opened port {tnc.location}')
                        self.__connections.append(tnc)
                    except Exception as error:
                        connection_errors.append(f'TNC - {error}')
                self.tncs = [
                    connection.location
                    for connection in self.__connections
                    if isinstance(connection, SerialTNC)
                    or isinstance(connection, RawAPRSTextFile)
                ]

                api_key = self.__configuration['aprs_fi']['aprs_fi_key']
                if api_key is None:
                    api_key = more_dialogs.ask_string(
                        'APRS.fi API Key',
                        'enter API key for https://aprs.fi \n'
                        'leave blank (or Cancel) for none',
                        parent=self.__windows['main'],
                    )
                try:
                    aprs_api = APRSfi(self.callsigns, api_key=api_key)
                    LOGGER.info(f'established connection to {aprs_api.location}')
                    self.__connections.append(aprs_api)
                    self.__configuration['aprs_fi']['aprs_fi_key'] = api_key
                except Exception as error:
                    connection_errors.append(f'aprs.fi - {error}')

                if (
                    'database' in self.__configuration
                    and self.__configuration['database']['database_hostname'] is not None
                ):
                    try:
                        ssh_tunnel_kwargs = {}
                        if 'ssh_tunnel' in self.__configuration:
                            ssh_hostname = self.__configuration['ssh_tunnel']['ssh_hostname']
                            if ssh_hostname is not None:
                                ssh_tunnel_kwargs.update(self.__configuration['ssh_tunnel'])
                                if '@' in ssh_hostname:
                                    (
                                        ssh_tunnel_kwargs['ssh_username'],
                                        ssh_tunnel_kwargs['ssh_hostname'],
                                    ) = ssh_hostname.split('@', 1)
                                if (
                                    'ssh_username' not in ssh_tunnel_kwargs
                                    or ssh_tunnel_kwargs['ssh_username'] is None
                                ):
                                    ssh_username = more_dialogs.ask_string(
                                        'SSH Tunnel Username',
                                        f'enter username for SSH host "{ssh_tunnel_kwargs["ssh_hostname"]}"',
                                        parent=self.__windows['main'],
                                    )
                                    if ssh_username is None or len(ssh_username) == 0:
                                        raise ConnectionError('missing SSH username')
                                    ssh_tunnel_kwargs['ssh_username'] = ssh_username

                                if (
                                    'ssh_password' not in ssh_tunnel_kwargs
                                    or ssh_tunnel_kwargs['ssh_password'] is None
                                ):
                                    password = more_dialogs.ask_string(
                                        'SSH Tunnel Password',
                                        f'enter password for SSH user '
                                        f'"{ssh_tunnel_kwargs["ssh_username"]}"',
                                        parent=self.__windows['main'],
                                    )
                                    if password is None or len(password) == 0:
                                        raise ConnectionError('missing SSH password')
                                    ssh_tunnel_kwargs['ssh_password'] = password

                        database_kwargs = {
                            key.replace('database_', ''): value
                            for key, value in self.__configuration['database'].items()
                        }
                        if (
                            'username' not in database_kwargs
                            or database_kwargs['username'] is None
                        ):
                            database_username = more_dialogs.ask_string(
                                'Database Username',
                                f'enter username for database '
                                f'"{database_kwargs["database_hostname"]}/'
                                f'{database_kwargs["database_database"]}"',
                                parent=self.__windows['main'],
                            )
                            if database_username is None or len(database_username) == 0:
                                raise ConnectionError('missing database username')
                            database_kwargs['username'] = database_username

                        if (
                            'password' not in database_kwargs
                            or database_kwargs['password'] is None
                        ):
                            database_password = more_dialogs.ask_string(
                                'Database Password',
                                f'enter password for database user '
                                f'"{database_kwargs["database_username"]}"',
                                parent=self.__windows['main'],
                            )
                            if database_password is None or len(database_password) == 0:
                                raise ConnectionError('missing database password')
                            database_kwargs['password'] = database_password
                        if 'table' not in database_kwargs or database_kwargs['table'] is None:
                            database_table = more_dialogs.ask_string(
                                'Database Table',
                                f'enter database table name',
                                parent=self.__windows['main'],
                            )
                            if database_table is None or len(database_table) == 0:
                                raise ConnectionError('missing database table name')
                            database_kwargs['table'] = database_table

                        self.database = APRSDatabaseTable(
                            **database_kwargs, **ssh_tunnel_kwargs, callsigns=self.callsigns
                        )
                        LOGGER.info(f'connected to {self.database.location}')
                        self.__connections.append(self.database)
                        self.__configuration['database'].update(database_kwargs)
                        self.__configuration['ssh_tunnel'].update(ssh_tunnel_kwargs)
                    except ConnectionError as error:
                        connection_errors.append(f'database - {error}')
                        self.database = None
                else:
                    self.database = None

                if self.igate:
                    try:
                        self.aprs_is = APRSis(self.callsigns)
                    except ConnectionError as error:
                        connection_errors.append(f'igate - {error}')
                        self.aprs_is = None

                if len(self.__connections) == 0:
                    if self.output_filename is not None and self.output_filename.exists():
                        self.__connections.append(PacketGeoJSON(self.output_filename))
                    else:
                        connection_errors = '\n'.join(connection_errors)
                        raise ConnectionError(f'no connections started\n{connection_errors}')

                LOGGER.info(
                    f'listening for packets every {self.interval_seconds}s from {len(self.__connections)} '
                    f'connection(s): {", ".join([connection.location for connection in self.__connections])}'
                )

                for variable, enabled in self.__plot_toggles.items():
                    enabled = enabled.get()
                    if enabled and variable not in self.__plots:
                        self.__plots[variable] = LivePlot(
                            self.packet_tracks, variable, self.predictions
                        )
                    elif not enabled and variable in self.__plots:
                        self.__plots[variable].close()
                        del self.__plots[variable]

                set_child_states(self.__frames['configuration'], 'disabled')

                for callsign in self.packet_tracks:
                    set_child_states(self.__windows[callsign], 'disabled')

                self.__running = True
            except Exception as error:
                teek.dialog.error(error.__class__.__name__, error)
                if '\n' in str(error):
                    for connection_error in str(error).split('\n'):
                        LOGGER.error(connection_error)
                else:
                    LOGGER.error(error)
                self.__running = False
                set_child_states(self.__frames['configuration'], 'normal')

            self.__elements['toggle_button'].busy_forget()
            asyncio.run(self.retrieve_packets())
        else:
            if self.__timeout is not None:
                try:
                    self.__timeout.cancel()
                except RuntimeError:
                    pass

            for connection in self.__connections:
                connection.close()

                if type(connection) is SerialTNC:
                    LOGGER.info(f'closing port {connection.location}')

            LOGGER.info(f'closed {len(self.__connections)} connections')

            for callsign in self.packet_tracks:
                set_child_states(self.__windows[callsign], 'disabled')
            set_child_states(self.__frames['configuration'], 'normal')

            if not self.toggles['log_file']:
                set_child_states(self.__elements['log_file_box'], 'disabled')
            if not self.toggles['output_file']:
                set_child_states(self.__elements['output_file_box'], 'disabled')
            if not self.toggles['prediction_file']:
                set_child_states(self.__elements['prediction_file_box'], 'disabled')

            self.__toggle_text.set('Start')
            self.__running = False
            self.__connections = []

            logging.shutdown()

    async def retrieve_packets(self):
        if self.running:
            try:
                current_time = datetime.now()

                new_packets = retrieve_packets(
                    self.__connections,
                    self.__packet_tracks,
                    self.database,
                    self.start_date,
                    self.end_date,
                    logger=LOGGER,
                )

                output_filename_index = None
                for index, connection in enumerate(self.__connections):
                    if isinstance(connection, PacketGeoJSON):
                        output_filename_index = index
                if output_filename_index is not None:
                    self.__connections.pop(output_filename_index)

                if self.toggles['prediction_file']:
                    try:
                        self.__predictions = get_predictions(
                            self.packet_tracks,
                            **{
                                key.replace('prediction_', ''): value
                                for key, value in self.__configuration['prediction'].items()
                                if 'prediction_' in key
                            },
                        )
                        if self.prediction_filename is not None:
                            write_packet_tracks(
                                self.__predictions.values(), self.prediction_filename
                            )
                    except PredictionError as error:
                        LOGGER.warning(f'{error.__class__.__name__} - {error}')
                    except Exception as error:
                        LOGGER.exception(
                            f'error retrieving prediction trajectory - {error.__class__.__name__} - {error}'
                        )

                for variable, plot in self.__plots.items():
                    plot.update(self.packet_tracks, self.predictions)

                if len(new_packets) > 0:
                    if self.output_filename is not None:
                        write_packet_tracks(
                            [
                                self.__packet_tracks[callsign]
                                for callsign in self.__packet_tracks
                            ],
                            self.output_filename,
                        )

                    await self.__update_sources_window(new_packets)

                    if self.aprs_is is not None:
                        for packets in new_packets.values():
                            self.aprs_is.send(packets)

                    updated_callsigns = {
                        packet.from_callsign
                        for packets in new_packets.values()
                        for packet in packets
                        if isinstance(packet, APRSPacket)
                    }
                else:
                    updated_callsigns = {}

                if len(self.__packet_tracks) > 0:
                    await asyncio.wait(
                        [
                            self.__update_callsign_window(
                                callsign,
                                only_time=callsign not in updated_callsigns,
                                current_time=current_time,
                            )
                            for callsign, packet_track in self.__packet_tracks.items()
                        ]
                    )

                if self.running:
                    self.__timeout = teek.after(
                        ms=int(self.interval_seconds * 1000),
                        callback=asyncio.run,
                        kwargs={'main': self.retrieve_packets()},
                    )
            except KeyboardInterrupt:
                self.close()
            except Exception as error:
                LOGGER.exception(f'{error.__class__.__name__} - {error}')

    def close(self):
        try:
            if self.running:
                self.toggle()
            for plot in self.__plots.values():
                plot.close()
            teek.quit()
        except Exception as error:
            LOGGER.exception(f'{error.__class__.__name__} - {error}')
        sys.exit()

    def __select_tnc(self):
        if self.__elements['tnc'].text == self.__file_selection_option:
            self.tncs = teek.dialog.open_multiple_files(
                title='Select TNC text file(s)...',
                defaultextension='.txt',
                filetypes=[('Text', '*.txt')],
            )

    def __select_log_file(self):
        self.log_filename = teek.dialog.save_file(
            title='Create log file...',
            initialdir=str(self.log_filename.parent),
            initialfile=self.log_filename.stem,
            defaultextension='.txt',
            filetypes=[('Text', '*.txt')],
        )

    def __select_output_file(self):
        self.output_filename = teek.dialog.save_file(
            title='Create output file...',
            initialdir=str(self.output_filename.parent),
            initialfile=self.output_filename.stem,
            defaultextension='.geojson',
            filetypes=[
                ('GeoJSON', '*.geojson'),
                ('Text', '*.txt'),
                ('Keyhole Markup Language', '*.kml'),
            ],
        )

    def __select_prediction_file(self):
        self.prediction_filename = teek.dialog.save_file(
            title='Create predict file...',
            initialdir=str(self.prediction_filename.parent),
            initialfile=self.prediction_filename.stem,
            defaultextension='.geojson',
            filetypes=[
                ('GeoJSON', '*.geojson'),
                ('Text', '*.txt'),
                ('Keyhole Markup Language', '*.kml'),
            ],
        )

    def __toggle_log_file(self, value: bool = None):
        if (value is not None and value) or self.toggles['log_file']:
            set_child_states(self.__elements['log_file_box'], state='normal')
            get_logger(LOGGER.name, log_filename=self.log_filename)
        else:
            set_child_states(self.__elements['log_file_box'], state='disabled')
            for existing_file_handler in [
                handler for handler in LOGGER.handlers if type(handler) is logging.FileHandler
            ]:
                LOGGER.removeHandler(existing_file_handler)

    def __toggle_output_file(self, value: bool = None):
        if (value is not None and value) or self.toggles['output_file']:
            set_child_states(self.__elements['output_file_box'], state='normal')
        else:
            set_child_states(self.__elements['output_file_box'], state='disabled')

    def __toggle_prediction_file(self, value: bool = None):
        if (value is not None and value) or self.toggles['prediction_file']:
            set_child_states(self.__elements['prediction_file_box'], state='normal')
        else:
            set_child_states(self.__elements['prediction_file_box'], state='disabled')

    def __add_callsign_window(self, callsign: str) -> teek.Window:
        window = teek.Window(callsign)

        self.__add_text_box(
            window, title=f'{callsign}.source', label=None, width=27, sticky='w', columnspan=3,
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.callsign',
            label='Callsign',
            width=17,
            sticky='w',
            columnspan=3,
        )
        self.__replace_text(self.__elements[f'{callsign}.callsign'], callsign)
        self.__add_text_box(
            window,
            title=f'{callsign}.packets',
            label='Packet #',
            width=17,
            sticky='w',
            columnspan=3,
        )

        self.__add_text_box(
            window,
            title=f'{callsign}.time',
            label='Time',
            width=19,
            sticky='w',
            row=self.__elements[f'{callsign}.source'].grid_info()['row'],
            column=self.__elements[f'{callsign}.source'].grid_info()['column'] + 4,
            columnspan=3,
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.age',
            label='Packet Age',
            width=14,
            units='s',
            sticky='w',
            row=self.__elements[f'{callsign}.callsign'].grid_info()['row'],
            column=self.__elements[f'{callsign}.callsign'].grid_info()['column'] + 3,
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.interval',
            label='Interval',
            width=14,
            units='s',
            sticky='w',
            row=self.__elements[f'{callsign}.packets'].grid_info()['row'],
            column=self.__elements[f'{callsign}.packets'].grid_info()['column'] + 3,
        )

        separator = teek.Separator(window, orient='horizontal')
        separator.grid(
            row=len(window.grid_rows), column=0, columnspan=7, sticky='ew', pady=10,
        )

        self.__add_text_box(
            window,
            title=f'{callsign}.coordinates',
            label='Lat., Lon.',
            width=17,
            sticky='w',
            columnspan=3,
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.distance',
            label='Distance',
            width=14,
            units='m',
            sticky='w',
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.ground_speed',
            label='Ground Speed',
            width=14,
            units='m/s',
            sticky='w',
        )

        self.__add_text_box(
            window,
            title=f'{callsign}.altitude',
            label='Alt.',
            width=14,
            units='m',
            sticky='w',
            row=self.__elements[f'{callsign}.coordinates'].grid_info()['row'],
            column=self.__elements[f'{callsign}.coordinates'].grid_info()['column'] + 3,
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.ascent',
            label='Ascent',
            width=14,
            units='m',
            sticky='w',
            row=self.__elements[f'{callsign}.distance'].grid_info()['row'],
            column=self.__elements[f'{callsign}.distance'].grid_info()['column'] + 3,
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.ascent_rate',
            label='Ascent Rate',
            width=14,
            units='m/s',
            sticky='w',
            row=self.__elements[f'{callsign}.ground_speed'].grid_info()['row'],
            column=self.__elements[f'{callsign}.ground_speed'].grid_info()['column'] + 3,
        )

        separator = teek.Separator(window, orient='horizontal')
        separator.grid(
            row=len(window.grid_rows), column=0, columnspan=7, sticky='ew', pady=10,
        )

        self.__add_text_box(
            window,
            title=f'{callsign}.distance_downrange',
            label='Downrange',
            width=14,
            units='m',
            sticky='w',
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.distance_overground',
            label='Overground',
            width=14,
            units='m',
            sticky='w',
        )

        self.__add_text_box(
            window,
            title=f'{callsign}.maximum_altitude',
            label='Max Alt.',
            width=14,
            units='m',
            sticky='w',
            row=self.__elements[f'{callsign}.distance_downrange'].grid_info()['row'],
            column=self.__elements[f'{callsign}.distance_downrange'].grid_info()['column'] + 3,
        )
        self.__add_text_box(
            window,
            title=f'{callsign}.time_to_ground',
            label='Est. Landing',
            width=14,
            units='s',
            sticky='w',
            row=self.__elements[f'{callsign}.distance_overground'].grid_info()['row'],
            column=self.__elements[f'{callsign}.distance_overground'].grid_info()['column']
            + 3,
        )

        separator = teek.Separator(window, orient='vertical')
        separator.grid(
            row=0, column=3, rowspan=len(window.grid_rows) + 2, sticky='ns', padx=10,
        )

        window.on_delete_window.connect(window.iconify)

        self.__windows[callsign] = window

        return window

    async def __update_callsign_window(
        self, callsign: str, only_time: bool = False, current_time: datetime = None
    ):
        if callsign not in self.__windows:
            self.__add_callsign_window(callsign)

        if current_time is None:
            current_time = datetime.now()

        window = self.__windows[callsign]

        if not only_time:
            window.deiconify()
            window.focus()

        set_child_states(window, 'normal')

        packet_track = self.packet_tracks[callsign]

        packet_time = datetime.utcfromtimestamp(
            (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00Z'))
            / numpy.timedelta64(1, 's')
        )

        if not only_time:
            self.__replace_text(self.__elements[f'{callsign}.packets'], len(packet_track))
            self.__replace_text(self.__elements[f'{callsign}.source'], packet_track[-1].source)
            self.__replace_text(self.__elements[f'{callsign}.time'], f'{packet_time}')
            self.__replace_text(
                self.__elements[f'{callsign}.altitude'],
                f'{packet_track.coordinates[-1, 2]:.3f}',
            )
            self.__replace_text(
                self.__elements[f'{callsign}.coordinates'],
                ', '.join(
                    f'{value:.3f}' for value in reversed(packet_track.coordinates[-1, :2])
                ),
            )
            self.__replace_text(
                self.__elements[f'{callsign}.ascent'], f'{packet_track.ascents[-1]:.2f}',
            )
            self.__replace_text(
                self.__elements[f'{callsign}.distance'],
                f'{packet_track.overground_distances[-1]:.2f}',
            )
            self.__replace_text(
                self.__elements[f'{callsign}.interval'], f'{packet_track.intervals[-1]:.2f}',
            )
            self.__replace_text(
                self.__elements[f'{callsign}.ascent_rate'],
                f'{packet_track.ascent_rates[-1]:.2f}',
            )
            self.__replace_text(
                self.__elements[f'{callsign}.ground_speed'],
                f'{packet_track.ground_speeds[-1]:.2f}',
            )

            self.__replace_text(
                self.__elements[f'{callsign}.distance_downrange'],
                f'{packet_track.distance_downrange:.2f}',
            )
            self.__replace_text(
                self.__elements[f'{callsign}.distance_overground'],
                f'{packet_track.length:.2f}',
            )

            self.__replace_text(
                self.__elements[f'{callsign}.maximum_altitude'],
                f'{packet_track.coordinates[:, 2].max():.2f}',
            )

        time_to_ground_box = self.__elements[f'{callsign}.time_to_ground']
        if packet_track.time_to_ground >= timedelta(seconds=0):
            time_to_ground_box.config['state'] = 'normal'
            current_time_to_ground = packet_time + packet_track.time_to_ground - current_time
            self.__replace_text(
                time_to_ground_box, f'{current_time_to_ground / timedelta(seconds=1):.2f}',
            )
        else:
            self.__replace_text(time_to_ground_box, '')
            time_to_ground_box.config['state'] = 'disabled'

        set_child_states(window, 'disabled', [teek.Text])

        packet_age_box = self.__elements[f'{callsign}.age']
        packet_age_box.config['state'] = 'normal'
        self.__replace_text(
            packet_age_box, f'{(current_time - packet_time) / timedelta(seconds=1):.2f}',
        )
        packet_age_box.config['state'] = 'disabled'

    def __add_sources_window(self) -> teek.Window:
        sources_window = teek.Window('sources')
        for index, connection in enumerate(self.__connections):
            self.__add_text_box(
                sources_window,
                title=f'sources.source_{index}_location',
                label=None,
                width=40,
                sticky='w',
            )
            self.__add_text_box(
                sources_window,
                title=f'sources.source_{index}_packets',
                label='packets',
                width=5,
                sticky='w',
                row=self.__elements[f'sources.source_{index}_location'].grid_info()['row'],
                column=self.__elements[f'sources.source_{index}_location'].grid_info()[
                    'column'
                ]
                + 1,
            )
            self.__replace_text(
                self.__elements[f'sources.source_{index}_location'], connection.location,
            )
        sources_window.on_delete_window.connect(sources_window.iconify)
        self.__windows['sources'] = sources_window

        for index, connection in enumerate(self.__connections):
            self.__replace_text(
                self.__elements[f'sources.source_{index}_packets'], 0,
            )

        return sources_window

    async def __update_sources_window(self, sources: {str: [LocationPacket]}):
        if 'sources' not in self.__windows:
            window = self.__add_sources_window()
        else:
            window = self.__windows['sources']

        for index, connection in enumerate(self.__connections):
            for source, packets in sources.items():
                if source == connection.location:
                    if f'sources.source_{index}_packets' not in self.__elements:
                        self.__add_text_box(
                            window,
                            title=f'sources.source_{index}_location',
                            label=None,
                            width=40,
                            sticky='w',
                        )
                        self.__add_text_box(
                            window,
                            title=f'sources.source_{index}_packets',
                            label='packets',
                            width=5,
                            sticky='w',
                            row=self.__elements[
                                f'sources.source_{index}_location'
                            ].grid_info()['row'],
                            column=self.__elements[
                                f'sources.source_{index}_location'
                            ].grid_info()['column']
                            + 1,
                        )
                        self.__replace_text(
                            self.__elements[f'sources.source_{index}_location'],
                            connection.location,
                        )
                        self.__replace_text(
                            self.__elements[f'sources.source_{index}_packets'], 0,
                        )
                    self.__replace_text(
                        self.__elements[f'sources.source_{index}_packets'],
                        int(self.__elements[f'sources.source_{index}_packets'].get())
                        + len(packets),
                    )

    def __add_combo_box(
        self,
        frame: teek.Frame,
        title: str,
        options: [str],
        option_select: Callable = None,
        **kwargs,
    ) -> teek.Combobox:
        width = kwargs['width'] if 'width' in kwargs else None
        combo_box = teek.Combobox(frame, width=width)
        combo_box.config['values'] = options
        if option_select is not None:
            combo_box.bind('<<ComboboxSelected>>', option_select)
        return self.__add_text_box(frame, title, text_box=combo_box, **kwargs)

    def __add_file_box(
        self, frame: teek.Frame, title: str, file_select: Callable, **kwargs
    ) -> teek.Frame:
        if 'row' not in kwargs:
            kwargs['row'] = len(frame.grid_rows)
        if 'column' not in kwargs:
            kwargs['column'] = 0
        file_box_kwargs = {
            key: value
            for key, value in kwargs.items()
            if key in ['row', 'column', 'columnspan']
        }
        if 'columnspan' in file_box_kwargs:
            file_box_kwargs['columnspan'] -= 1

        if 'label' in kwargs:
            text_label = teek.Label(frame, text=kwargs['label'])
            text_label.grid(row=kwargs['row'], column=kwargs['column'], sticky='w')
            file_box_kwargs['column'] += 1

        file_box_frame = teek.Frame(frame)
        file_box_frame.grid(**file_box_kwargs)

        file_box = teek.Entry(
            file_box_frame, width=kwargs['width'] if 'width' in kwargs else None
        )
        file_button = teek.Button(file_box_frame, text='...', command=file_select)

        file_box.pack(side='left')
        file_button.pack(side='left')

        self.__elements[title] = file_box
        return file_box_frame

    def __add_entry_box(self, frame: teek.Frame, title: str, **kwargs) -> teek.Entry:
        if 'text_box' not in kwargs:
            kwargs['text_box'] = teek.Entry(
                frame, width=kwargs['width'] if 'width' in kwargs else None
            )
        return self.__add_text_box(frame, title, **kwargs)

    def __add_text_box(
        self,
        frame: teek.Frame,
        title: str,
        label: str,
        units: str = None,
        row: int = None,
        column: int = None,
        width: int = 10,
        text_box: teek.Entry = None,
        **kwargs,
    ) -> teek.Text:
        if row is None:
            row = len(frame.grid_rows)
        if column is None:
            column = 0

        if 'columnspan' in kwargs:
            if label is not None:
                kwargs['columnspan'] -= 1
            if units is not None:
                kwargs['columnspan'] -= 1

        if label is not None:
            text_label = teek.Label(frame, text=label)
            text_label.grid(row=row, column=column, sticky='w')
            column += 1

        if text_box is None:
            text_box = teek.Text(frame, width=width, height=1)
        text_box.grid(row=row, column=column, **kwargs)
        column += 1

        if units is not None:
            units_label = teek.Label(frame, text=units)
            units_label.grid(row=row, column=column)
            column += 1

        self.__elements[title] = text_box
        return text_box

    @staticmethod
    def __replace_text(element: teek.Entry, value: str):
        if value is None:
            value = ''

        if isinstance(element, teek.Text):
            element.delete(element.start, element.end)
            element.insert(element.start, str(value))
        elif isinstance(element, teek.Entry):
            element.text = str(value)


def set_child_states(frame: teek.Frame, state: str = None, types: [type] = None):
    if state is None:
        state = 'normal'
    for child in frame.winfo_children():
        if isinstance(child, teek.Frame):
            set_child_states(child, state, types)
        else:
            if types is None or any(
                isinstance(child, selected_type) for selected_type in types
            ):
                try:
                    child.config['state'] = state
                except teek.TclError:
                    continue
                except KeyError:
                    if state == 'normal':
                        child.state.discard('disabled')
                    else:
                        child.state.add(state)



from datetime import datetime, timedelta
from typing import Iterable, Union

from dateutil.parser import parse as parse_date
import numpy
from pandas import DataFrame
from pyproj import CRS

from packetraven.model import (
    FREEFALL_DESCENT_RATE,
    FREEFALL_DESCENT_RATE_UNCERTAINTY,
    FREEFALL_SECONDS_TO_GROUND,
)
from packetraven.packets import APRSPacket, DEFAULT_CRS, LocationPacket
from packetraven.packets.structures import DoublyLinkedList


class LocationPacketTrack:
    """ collection of location packets """

    def __init__(self, name: str, packets: [LocationPacket] = None, crs: CRS = None):
        """
        location packet track

        :param name: name of packet track
        :param crs: coordinate reference system to use
        :param packets: iterable of packets
        """

        self.name = name
        self.packets = DoublyLinkedList(None)
        self.crs = crs if crs is not None else DEFAULT_CRS

        if packets is not None:
            for packet in packets:
                self.append(packet)

    def append(self, packet: LocationPacket):
        if packet not in self.packets:
            if packet.crs != self.crs:
                packet.transform_to(self.crs)
            self.packets.append(packet)

    def extend(self, packets: [LocationPacket]):
        for packet in packets:
            self.append(packet)

    def sort(self):
        self.packets.sort()

    @property
    def times(self) -> numpy.ndarray:
        return numpy.array([packet.time for packet in self.packets], dtype=numpy.datetime64)

    @property
    def coordinates(self) -> numpy.ndarray:
        return numpy.stack([packet.coordinates for packet in self.packets], axis=0)

    @property
    def altitudes(self) -> numpy.ndarray:
        return self.coordinates[:, 2]

    @property
    def intervals(self) -> numpy.ndarray:
        return numpy.concatenate(
            [
                [0],
                numpy.array(
                    [packet_delta.seconds for packet_delta in numpy.diff(self.packets)]
                ),
            ]
        )

    @property
    def overground_distances(self) -> numpy.ndarray:
        """ overground distances between packets """
        return numpy.concatenate(
            [
                [0],
                numpy.array(
                    [packet_delta.overground for packet_delta in self.packets.difference]
                ),
            ]
        )

    @property
    def ascents(self) -> numpy.ndarray:
        """ differences in altitude between packets """
        return numpy.concatenate(
            [
                [0],
                numpy.array(
                    [packet_delta.ascent for packet_delta in numpy.diff(self.packets)]
                ),
            ]
        )

    @property
    def ascent_rates(self) -> numpy.ndarray:
        """ instantaneous ascent rates between packets """
        return numpy.concatenate(
            [
                [0],
                numpy.array(
                    [packet_delta.ascent_rate for packet_delta in numpy.diff(self.packets)]
                ),
            ]
        )

    @property
    def ground_speeds(self) -> numpy.ndarray:
        """ instantaneous overground speeds between packets """
        return numpy.concatenate(
            [
                [0],
                numpy.array(
                    [packet_delta.ground_speed for packet_delta in numpy.diff(self.packets)]
                ),
            ]
        )

    @property
    def cumulative_overground_distances(self) -> numpy.ndarray:
        """ cumulative overground distances from start """
        return numpy.cumsum(self.overground_distances)

    @property
    def time_to_ground(self) -> timedelta:
        """ estimated time to reach the ground at the current rate of descent """

        current_ascent_rate = self.ascent_rates[-1]
        if current_ascent_rate < 0:
            # TODO implement landing location as the intersection of the predicted descent track with a local DEM
            # TODO implement a time to impact calc based off of standard atmo
            return timedelta(seconds=self.altitudes[-1] / abs(current_ascent_rate))
        else:
            return timedelta(seconds=-1)

    @property
    def distance_downrange(self) -> float:
        """ direct overground distance between first and last packets only """
        if len(self.packets) > 0:
            return self.packets[-1].overground_distance(self.coordinates[0, :2])
        else:
            return 0.0

    @property
    def length(self) -> float:
        """ total length of the packet track over the ground """
        return sum([distance.overground for distance in self.packets.difference])

    def __getitem__(
        self, index: Union[int, Iterable[int], slice]
    ) -> Union[LocationPacket, 'LocationPacketTrack']:
        if isinstance(index, str):
            try:
                index = parse_date(index)
            except:
                index = int(index)
        if isinstance(index, int):
            return self.packets[index]
        elif isinstance(index, Iterable) or isinstance(index, slice):
            if isinstance(index, numpy.ndarray) and index.dtype == bool:
                index = numpy.where(index)[0]
            if isinstance(index, slice) or len(index) > 0:
                packets = self.packets[index]
            else:
                packets = None
            return self.__class__(self.name, packets, self.crs)
        elif isinstance(index, datetime):
            matching_packets = []
            for packet in self.packets:
                if packet.time == index:
                    matching_packets.append(packet)
            if len(matching_packets) == 0:
                maximum_interval = numpy.timedelta64(int(max(self.intervals)), 's')
                if (
                    index > min(self.times) - maximum_interval
                    and index < max(self.times) + maximum_interval
                ):
                    smallest_difference = None
                    closest_time = None
                    for packet in self.packets:
                        difference = abs(packet.time - index)
                        if smallest_difference is None or difference < smallest_difference:
                            smallest_difference = difference
                            closest_time = packet.time
                    return self[closest_time]
                else:
                    raise IndexError(f'time index out of range: {index}')
            elif len(matching_packets) == 1:
                return matching_packets[0]
            else:
                return self.__class__(self.name, matching_packets, self.crs)
        else:
            raise ValueError(f'unrecognized index: {index}')

    def __iter__(self):
        return iter(self.packets)

    def __reversed__(self):
        return reversed(self.packets)

    def __contains__(self, item) -> bool:
        return item in self.packets

    def __len__(self) -> int:
        return len(self.packets)

    def __eq__(self, other) -> bool:
        return self.packets == other.packets

    def __str__(self) -> str:
        return str(list(self))

    @property
    def dataframe(self) -> DataFrame:
        return DataFrame(
            {
                'name': [self.name for _ in range(len(self))],
                'times': self.times,
                'x': self.coordinates[:, 0],
                'y': self.coordinates[:, 1],
                'z': self.coordinates[:, 2],
                'intervals': self.intervals,
                'overground_distances': self.overground_distances,
                'ascents': self.ascents,
                'ascent_rates': self.ascent_rates,
                'ground_speeds': self.ground_speeds,
                'cumulative_overground_distances': self.cumulative_overground_distances,
            }
        )


class BalloonTrack(LocationPacketTrack):
    def __init__(self, name: str, packets: [LocationPacket] = None, crs: CRS = None):
        super().__init__(name, packets, crs)
        self.__falling = False

    @property
    def time_to_ground(self) -> timedelta:
        current_ascent_rate = self.ascent_rates[-1]
        if current_ascent_rate < 0:
            if self.falling:
                return timedelta(seconds=FREEFALL_SECONDS_TO_GROUND(self.altitudes[-1]))
            else:
                current_altitude = self.altitudes[-1]
                # TODO implement landing location as the intersection of the predicted descent track with a local DEM
                return timedelta(seconds=current_altitude / -current_ascent_rate)
        else:
            return timedelta(seconds=-1)

    @property
    def falling(self) -> bool:
        current_ascent_rate = self.ascent_rates[-1]
        if current_ascent_rate >= 0:
            self.__falling = False
        elif not self.__falling:
            current_altitude = self.altitudes[-1]
            freefall_descent_rate = FREEFALL_DESCENT_RATE(current_altitude)
            freefall_descent_rate_uncertainty = FREEFALL_DESCENT_RATE_UNCERTAINTY(
                current_altitude
            )
            if numpy.abs(current_ascent_rate - freefall_descent_rate) < numpy.abs(
                freefall_descent_rate_uncertainty
            ):
                self.__falling = True
        return self.__falling


class APRSTrack(BalloonTrack):
    """ collection of APRS location packets """

    def __init__(self, callsign: str, packets: [APRSPacket] = None, crs: CRS = None):
        """
        APRS packet track

        :param callsign: callsign of APRS packets
        :param packets: iterable of packets
        :param crs: coordinate reference system to use
        """

        if not isinstance(callsign, str):
            callsign = str(callsign)
        if len(callsign) > 9 or ' ' in callsign:
            raise ValueError(f'unrecognized callsign format: "{callsign}"')

        self.__callsign = callsign
        super().__init__(self.callsign, packets, crs)

    @property
    def callsign(self) -> str:
        return self.__callsign

    def append(self, packet: APRSPacket):
        packet_callsign = packet['callsign']

        if packet_callsign == self.callsign:
            super().append(packet)
        else:
            raise ValueError(
                f'Packet callsign {packet_callsign} does not match ground track callsign {self.callsign}.'
            )

    def __eq__(self, other) -> bool:
        return self.callsign == other.__callsign and super().__eq__(other)

    def __str__(self) -> str:
        return f'{self.callsign}: {super().__str__()}'


class PredictedTrajectory(LocationPacketTrack):
    def __init__(
        self, name: str, packets: [LocationPacket], prediction_time: datetime, crs: CRS = None
    ):
        """
        prediction trajectory

        :param name: name of prediction
        :param packets: iterable of packets
        :param prediction_time: time of prediction
        :param crs: coordinate reference system to use
        """

        if isinstance(prediction_time, str):
            prediction_time = parse_date(prediction_time)

        super().__init__(name, packets, crs)
        self.prediction_time = prediction_time

    @property
    def landing_site(self) -> (float, float, float):
        return self.coordinates[-1]



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Generator, Optional, Tuple, Union

from dateutil.parser import parse as parse_date
import numpy
import pytz
import requests
from shapely.geometry import Point

from packetraven.model import FREEFALL_DESCENT_RATE
from packetraven.packets import LocationPacket
from packetraven.packets.tracks import LocationPacketTrack, PredictedTrajectory
from packetraven.utilities import get_logger

DEFAULT_FLOAT_ALTITUDE_UNCERTAINTY = 500
UTC_TIMEZONE = pytz.utc

LOGGER = get_logger('predicts')


class PredictionAPIURL(Enum):
    cusf = 'https://predict.cusf.co.uk/api/v1/'
    lukerenegar = 'https://predict.lukerenegar.com/api/v1.1/'


class FlightProfile(Enum):
    standard = 'standard_profile'
    float = 'float_profile'


class BalloonPredictionQuery(ABC):
    def __init__(
        self,
        api_url: str,
        launch_site: Union[Tuple[float, float, Optional[float]], Point],
        launch_time: datetime,
        ascent_rate: float,
        burst_altitude: float,
        sea_level_descent_rate: float,
        float_altitude: float = None,
        float_end_time: datetime = None,
        name: str = None,
        descent_only: bool = False,
    ):
        """
        initialize a new balloon prediction API query

        :param api_url: URL of API
        :param launch_site: location of balloon launch
        :param launch_time: date and time of balloon launch
        :param ascent_rate: average ascent rate (m/s)
        :param burst_altitude: altitude at which balloon will burst
        :param sea_level_descent_rate: descent rate at sea level (m/s)
        :param float_altitude: altitude of float (m)
        :param float_end_time: date and time of float end
        :param name: name of prediction track
        :param descent_only: whether to query for descent only
        """

        if not isinstance(launch_site, Point):
            launch_site = Point(*launch_site)

        if name is None:
            name = 'prediction'

        if launch_time is not None:
            if launch_time.tzinfo is None or launch_time.tzinfo.utcoffset(launch_time) is None:
                launch_time = UTC_TIMEZONE.localize(launch_time)

        if float_end_time is not None:
            if (
                float_end_time.tzinfo is None
                or float_end_time.tzinfo.utcoffset(float_end_time) is None
            ):
                float_end_time = UTC_TIMEZONE.localize(float_end_time)

        self.api_url = api_url
        self.launch_site = launch_site
        self.launch_time = launch_time
        self.ascent_rate = ascent_rate
        self.burst_altitude = burst_altitude
        self.sea_level_descent_rate = abs(sea_level_descent_rate)
        self.float_altitude = float_altitude
        self.float_end_time = float_end_time
        self.name = name
        self.descent_only = descent_only

    @property
    @abstractmethod
    def query(self) -> {str: Any}:
        raise NotImplementedError

    def get(self) -> {str: Any}:
        response = requests.get(self.api_url, params=self.query)
        return response.json()

    @property
    @abstractmethod
    def predict(self) -> PredictedTrajectory:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.api_url)}, {repr(self.launch_site)}, {repr(self.launch_time)}, {repr(self.ascent_rate)}, {repr(self.burst_altitude)}, {repr(self.sea_level_descent_rate)})'


class PredictionError(Exception):
    pass


class CUSFBalloonPredictionQuery(BalloonPredictionQuery):
    def __init__(
        self,
        launch_site: Union[Tuple[float, float], Point],
        launch_time: datetime,
        ascent_rate: float,
        burst_altitude: float,
        sea_level_descent_rate: float,
        profile: FlightProfile = None,
        version: float = None,
        dataset_time: datetime = None,
        float_altitude: float = None,
        float_end_time: datetime = None,
        api_url: PredictionAPIURL = None,
        name: str = None,
        descent_only: bool = False,
    ):
        if profile is None:
            if float_altitude is not None or float_end_time is not None:
                profile = FlightProfile.float
            else:
                profile = FlightProfile.standard

        if dataset_time is not None:
            if (
                dataset_time.tzinfo is None
                or dataset_time.tzinfo.utcoffset(dataset_time) is None
            ):
                dataset_time = UTC_TIMEZONE.localize(dataset_time)

        if api_url is None:
            api_url = PredictionAPIURL.cusf

        if isinstance(api_url, PredictionAPIURL):
            api_url = api_url.value

        if name is None:
            name = 'cusf_prediction'

        super().__init__(
            api_url,
            launch_site,
            launch_time,
            ascent_rate,
            burst_altitude,
            sea_level_descent_rate,
            float_altitude,
            float_end_time,
            name,
            descent_only,
        )

        # CUSF API requires longitude in 0-360 format
        if self.launch_site.x < 0:
            launch_coordinates = [self.launch_site.x + 360, self.launch_site.y]
            if self.launch_site.has_z:
                launch_coordinates.append(self.launch_site.z)
            self.launch_site = Point(launch_coordinates)

        self.profile = profile if not self.descent_only else FlightProfile.standard
        self.version = version
        self.dataset_time = dataset_time

    @property
    def query(self) -> {str: Any}:
        query = {
            'launch_longitude': self.launch_site.x,
            'launch_latitude': self.launch_site.y,
            'launch_datetime': self.launch_time.isoformat(),
            'ascent_rate': self.ascent_rate,
            'burst_altitude': self.burst_altitude,
            'descent_rate': self.sea_level_descent_rate,
        }

        if self.launch_site.has_z:
            query['launch_altitude'] = self.launch_site.z
        if self.profile is not None:
            query['profile'] = self.profile.value
        if self.version is not None:
            query['version'] = self.version
        if self.dataset_time is not None:
            query['dataset'] = self.dataset_time.isoformat()

        if self.profile == FlightProfile.float:
            if self.float_altitude is None:
                self.float_altitude = self.burst_altitude
            if self.float_end_time is None:
                raise PredictionError('float stop time `float_end_time` not provided')
            query['float_altitude'] = self.float_altitude
            query['stop_datetime'] = self.float_end_time.isoformat()

        return query

    def get(self) -> {str: Any}:
        response = requests.get(self.api_url, params=self.query)

        if response.status_code == 200:
            response = response.json()
            if 'error' not in response:
                # TODO tawhiri currently does not include descent when querying a float profile
                if self.profile == FlightProfile.float:
                    # this code runs another prediction query with a standard profile and extracts the descent stage to append to the response from the original query
                    for stage in response['prediction']:
                        # if a descent stage exists, we don't need to do anything
                        if stage['stage'] == 'descent':
                            break
                    else:
                        for stage in response['prediction']:
                            if stage['stage'] == 'float':
                                float_end = stage['trajectory'][-1]
                                break
                        else:
                            raise PredictionError('API did not return a float trajectory')

                        standard_profile_query = self.__class__(
                            launch_site=[
                                float_end['longitude'],
                                float_end['latitude'],
                                float_end['altitude'],
                            ],
                            launch_time=parse_date(float_end['datetime']),
                            ascent_rate=10,
                            burst_altitude=float_end['altitude'] + 0.1,
                            sea_level_descent_rate=self.sea_level_descent_rate,
                            profile=FlightProfile.standard,
                            version=self.version,
                            dataset_time=self.dataset_time,
                            float_altitude=None,
                            float_end_time=None,
                            api_url=self.api_url,
                            name=self.name,
                        )

                        for stage in standard_profile_query.get()['prediction']:
                            if stage['stage'] == 'descent':
                                response['prediction'].append(stage)
                                break

                if self.descent_only:
                    indices_to_remove = []
                    for index, stage in enumerate(response['prediction']):
                        # if a descent stage exists, we don't need to do anything
                        if stage['stage'] != 'descent':
                            indices_to_remove.append(index)
                            break
                    for index in indices_to_remove:
                        response['prediction'].pop(index)

                return response
            else:
                raise PredictionError(response['error']['description'])
        else:
            raise ConnectionError(f'connection raised error {response.status_code}')

    @property
    def predict(self) -> PredictedTrajectory:
        response = self.get()

        if 'error' not in response:
            points = []

            for stage in response['prediction']:
                points.extend(stage['trajectory'])

            for point in points:
                if point['longitude'] > 180:
                    point['longitude'] -= 360

            return PredictedTrajectory(
                name=self.name,
                packets=[
                    LocationPacket(
                        point['datetime'],
                        point['longitude'],
                        point['latitude'],
                        point['altitude'],
                    )
                    for point in points
                ],
                prediction_time=response['metadata']['complete_datetime'],
            )
        else:
            raise PredictionError(response['error']['description'])


class LukeRenegarBalloonPredictionQuery(CUSFBalloonPredictionQuery):
    def __init__(
        self,
        launch_site: Union[Tuple[float, float], Point],
        launch_time: datetime,
        ascent_rate: float,
        burst_altitude: float,
        sea_level_descent_rate: float,
        ascent_rate_standard_deviation: float = None,
        burst_altitude_standard_deviation: float = None,
        descent_rate_standard_deviation: float = None,
        wind_standard_deviation: float = None,
        use_monte_carlo: bool = None,
        physics_model: str = None,
        profile: FlightProfile = None,
        version: float = None,
        dataset_time: datetime = None,
        float_altitude: float = None,
        float_end_time: datetime = None,
        api_url: str = None,
        name: str = None,
        descent_only: bool = False,
    ):
        if api_url is None:
            api_url = PredictionAPIURL.lukerenegar

        if name is None:
            name = 'lrenegar_prediction'

        super().__init__(
            launch_site,
            launch_time,
            ascent_rate,
            burst_altitude,
            sea_level_descent_rate,
            profile,
            version,
            dataset_time,
            float_altitude,
            float_end_time,
            api_url,
            name,
            descent_only,
        )

        self.ascent_rate_standard_deviation = ascent_rate_standard_deviation
        self.burst_altitude_standard_deviation = burst_altitude_standard_deviation
        self.descent_rate_standard_deviation = descent_rate_standard_deviation
        self.wind_standard_deviation = wind_standard_deviation
        self.use_monte_carlo = use_monte_carlo
        self.physics_model = physics_model

    @property
    def query(self):
        query = super().query

        if self.ascent_rate_standard_deviation is not None:
            query['ascent_rate_std_dev'] = self.ascent_rate_standard_deviation
        if self.burst_altitude_standard_deviation is not None:
            query['burst_altitude_std_dev'] = self.burst_altitude_standard_deviation
        if self.descent_rate_standard_deviation is not None:
            query['descent_rate_std_dev'] = self.descent_rate_standard_deviation
        if self.wind_standard_deviation is not None:
            query['wind_std_dev'] = self.wind_standard_deviation
        if self.use_monte_carlo is not None:
            query['monte_carlo'] = self.use_monte_carlo
        if self.physics_model is not None:
            query['physics_model'] = self.physics_model

        return query


def get_predictions(
    packet_tracks: {str: LocationPacketTrack},
    start_location: (float, float, float) = None,
    start_time: datetime = None,
    ascent_rate: float = None,
    burst_altitude: float = None,
    sea_level_descent_rate: float = None,
    float_altitude: float = None,
    float_altitude_uncertainty: float = None,
    float_duration: timedelta = None,
    api_url: str = None,
) -> [PredictedTrajectory]:
    """
    Return location tracks detailing predicted trajectory of balloon flight(s) from current location.

    :param packet_tracks: location packet tracks
    :param start_time: start location
    :param start_location: start time
    :param ascent_rate: ascent rate (m/s)
    :param burst_altitude: altitude at which balloon will burst (m)
    :param sea_level_descent_rate: descent rate of payload at sea level (m/s)
    :param float_altitude: altitude at which to float (m)
    :param float_altitude_uncertainty: tolerance around which to consider the balloon "at float altitude"
    :param float_duration: expected duration of float
    :param api_url: URL of prediction API to use
    """

    if start_location is not None:
        if isinstance(start_location, Generator):
            start_location = list(start_location)
        if len(start_location) == 0:
            start_location = None
        elif len(start_location) == 2:
            start_location = (*start_location, 0)

    if float_altitude is not None and float_duration is None:
        raise ValueError('`float_duration` was not provided')

    if float_duration is not None and float_altitude is None:
        float_altitude = burst_altitude

    if float_altitude_uncertainty is None:
        float_altitude_uncertainty = DEFAULT_FLOAT_ALTITUDE_UNCERTAINTY

    if api_url is None:
        api_url = PredictionAPIURL.cusf

    prediction_tracks = {}
    for name, packet_track in packet_tracks.items():
        prediction_ascent_rate = None
        prediction_burst_altitude = None

        ascent_rates = packet_track.ascent_rates[packet_track.ascent_rates > 0]
        if len(ascent_rates) > 0:
            average_ascent_rate = numpy.mean(ascent_rates)
            if average_ascent_rate > 0:
                prediction_ascent_rate = average_ascent_rate

            # if packet track is descending, override given burst altitude
            if len(ascent_rates) > 2 and all(ascent_rates[-2:] < 0):
                prediction_burst_altitude = packet_track.altitudes[-1] + 1

        if prediction_ascent_rate is None:
            prediction_ascent_rate = ascent_rate

        if prediction_burst_altitude is None:
            prediction_burst_altitude = burst_altitude

        if sea_level_descent_rate is not None:
            prediction_sea_level_descent_rate = sea_level_descent_rate
        else:
            prediction_sea_level_descent_rate = FREEFALL_DESCENT_RATE(0)

        prediction_start_time = packet_track[-1].time

        try:
            prediction_start_location = packet_track[prediction_start_time].coordinates
            if len(prediction_start_location.shape) > 1:
                prediction_start_location = prediction_start_location[0, :]
        except KeyError:
            prediction_start_location = packet_track[-1].coordinates
        if len(prediction_start_location) == 2:
            prediction_start_location = (*prediction_start_location, 0)

        if float_altitude is not None and not packet_track.falling:
            packets_at_float_altitude = packet_track[
                numpy.abs(float_altitude - packet_track.altitudes) < float_altitude_uncertainty
            ]
            if (
                len(packets_at_float_altitude) > 0
                and packets_at_float_altitude[-1].time == packet_track.times[-1]
            ):
                prediction_float_start_time = packets_at_float_altitude[0].time
                descent_only = False
            elif packet_track.ascent_rates[-1] >= 0:
                prediction_float_start_time = prediction_start_time + timedelta(
                    seconds=(float_altitude - prediction_start_location[2])
                    / prediction_ascent_rate
                )
                descent_only = False
            else:
                prediction_float_start_time = None
                descent_only = True
                prediction_sea_level_descent_rate = packet_track.ascent_rates[-1]
            if prediction_float_start_time is not None:
                prediction_float_end_time = prediction_float_start_time + float_duration
            else:
                prediction_float_end_time = None
        else:
            prediction_float_end_time = None
            descent_only = packet_track.falling or packet_track.ascent_rates[-1] < 0

        try:
            prediction_query = CUSFBalloonPredictionQuery(
                launch_site=prediction_start_location,
                launch_time=prediction_start_time,
                ascent_rate=prediction_ascent_rate,
                burst_altitude=prediction_burst_altitude,
                sea_level_descent_rate=prediction_sea_level_descent_rate,
                float_altitude=float_altitude,
                float_end_time=prediction_float_end_time,
                api_url=api_url,
                name=name,
                descent_only=descent_only,
            )

            prediction = prediction_query.predict

            if packet_track.time_to_ground >= timedelta(seconds=0):
                LOGGER.info(
                    f'{packet_track.name:<9} - predicted landing location: {prediction.coordinates[-1]} - https://www.google.com/maps/place/{prediction.coordinates[-1][1]},{prediction.coordinates[-1][0]}'
                )

            prediction_tracks[name] = prediction
        except Exception as error:
            LOGGER.warning(
                f'error retrieving prediction trajectory - {error.__class__.__name__} - {error}'
            )

    if all(
        value is not None
        for value in [
            start_location,
            start_time,
            ascent_rate,
            burst_altitude,
            sea_level_descent_rate,
        ]
    ):
        name = 'flight'

        if all(value is not None for value in [float_altitude, float_duration]):
            float_end_time = (
                start_time
                + timedelta(seconds=(float_altitude - start_location[2]) / ascent_rate)
                + float_duration
            )
        else:
            float_end_time = None

        try:
            prediction_query = CUSFBalloonPredictionQuery(
                launch_site=start_location,
                launch_time=start_time,
                ascent_rate=ascent_rate,
                burst_altitude=burst_altitude,
                sea_level_descent_rate=sea_level_descent_rate,
                float_altitude=float_altitude,
                float_end_time=float_end_time,
                api_url=api_url,
                name=name,
                descent_only=False,
            )

            prediction = prediction_query.predict
            prediction_tracks[name] = prediction
        except Exception as error:
            LOGGER.warning(
                f'error retrieving prediction trajectory - {error.__class__.__name__} - {error}'
            )

    return prediction_tracks



from datetime import datetime, timedelta
from time import sleep
from typing import Any, Sequence

import aprslib
import requests
from shapely.geometry import Point
from tablecrow import PostGresTable
from tablecrow.utilities import split_hostname_port

from packetraven.connections.base import (
    APRSPacketSink,
    APRSPacketSource,
    CREDENTIALS_FILENAME,
    LOGGER,
    NetworkConnection,
    PacketSink,
    PacketSource,
    TimeIntervalError,
)
from packetraven.packets import APRSPacket, LocationPacket
from packetraven.packets.parsing import InvalidPacketError
from packetraven.utilities import read_configuration


class APRSfi(APRSPacketSource, NetworkConnection):
    interval = timedelta(seconds=10)

    def __init__(self, callsigns: [str], api_key: str = None):
        """
        connect to https://aprs.fi

        :param callsigns: list of callsigns to return from source
        :param api_key: API key for aprs.fi
        """

        url = 'https://api.aprs.fi/api/get'
        if callsigns is None:
            raise ConnectionError(f'queries to {url} require a list of callsigns')
        super().__init__(url, callsigns)

        if api_key is None or api_key == '':
            configuration = read_configuration(CREDENTIALS_FILENAME)

            if 'APRS_FI' in configuration:
                api_key = configuration['APRS_FI']['api_key']
            else:
                raise ConnectionError(f'no APRS.fi API key specified')

        if not self.connected:
            raise ConnectionError(f'no network connection')

        self.api_key = api_key

        self.__last_access_time = None

    @property
    def api_key(self) -> str:
        return self.__api_key

    @api_key.setter
    def api_key(self, api_key: str):
        response = requests.get(
            f'{self.location}?name=OH2TI&what=wx&apikey={api_key}&format=json'
        ).json()
        if response['result'] == 'fail':
            raise ConnectionError(response['description'])
        self.__api_key = api_key

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        query = {
            'name': ','.join(self.callsigns),
            'what': 'loc',
            'apikey': self.api_key,
            'format': 'json',
            'timerange': int(timedelta(days=1) / timedelta(seconds=1)),
            'tail': int(timedelta(days=1) / timedelta(seconds=1)),
        }

        query = '&'.join(f'{key}={value}' for key, value in query.items())

        response = requests.get(f'{self.location}?{query}').json()
        if response['result'] != 'fail':
            packets = []
            for packet_candidate in response['entries']:
                try:
                    packet = APRSPacket.from_frame(packet_candidate, source=self.location)
                    packets.append(packet)
                except Exception as error:
                    LOGGER.error(f'{error.__class__.__name__} - {error}')
        else:
            LOGGER.warning(f'query failure "{response["code"]}: {response["description"]}"')
            packets = []

        self.__last_access_time = datetime.now()
        return packets

    def close(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.callsigns)}, {repr("****")})'


class PacketDatabaseTable(PostGresTable, PacketSource, PacketSink):
    __default_fields = {
        'time': datetime,
        'x': float,
        'y': float,
        'z': float,
        'source': str,
        'point': Point,
    }

    def __init__(self, hostname: str, database: str, table: str, **kwargs):
        if 'primary_key' not in kwargs:
            kwargs['primary_key'] = 'time'
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'] = {**self.__default_fields, **kwargs['fields']}
        PostGresTable.__init__(
            self, hostname=hostname, database=database, table_name=table, **kwargs
        )
        PacketSource.__init__(
            self, f'postgresql://{self.hostname}:{self.port}/{self.database}/{self.name}'
        )

        if not self.connected:
            raise ConnectionError(f'cannot connect to {self.location}')

        self.__last_access_time = None
        self.__send_buffer = []

    @property
    def packets(self) -> [LocationPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )
        packets = [
            LocationPacket(**{key: value for key, value in record.items() if key != 'point'})
            for record in self.records
        ]
        self.__last_access_time = datetime.now()
        return packets

    def __getitem__(self, key: Any) -> LocationPacket:
        record = super().__getitem__(key)
        record = {key.replace('packet_', ''): value for key, value in record.items()}
        return LocationPacket(**record, crs=self.crs)

    def __setitem__(self, time: datetime, packet: LocationPacket):
        if packet.crs != self.crs:
            packet.transform_to(self.crs)
        PostGresTable.__setitem__(self, time, self.__packet_record(packet))

    def insert(self, packets: [APRSPacket]):
        for packet in packets:
            if packet.crs != self.crs:
                packet.transform_to(self.crs)
        records = [self.__packet_record(packet) for packet in packets if packet not in self]
        PostGresTable.insert(self, records)

    def __contains__(self, packet: LocationPacket) -> bool:
        if isinstance(packet, LocationPacket):
            packet = [packet[key.replace('packet_', '')] for key in self.primary_key]
        return PostGresTable.__contains__(self, packet)

    def send(self, packets: [LocationPacket]):
        new_packets = [packet for packet in packets if packet not in self]
        if len(self.__send_buffer) > 0:
            new_packets.extend(self.__send_buffer)
            self.__send_buffer.clear()
        if len(new_packets) > 0:
            LOGGER.info(f'sending {len(new_packets)} packet(s) to {self.location}')
            try:
                self.insert(new_packets)
            except ConnectionError as error:
                LOGGER.info(
                    f'could not send packet(s) ({error}); reattempting on next iteration'
                )
                self.__send_buffer.extend(new_packets)

    def close(self):
        self.connection.close()
        if self.tunnel is not None:
            self.tunnel.stop()

    @staticmethod
    def __packet_record(packet: LocationPacket) -> {str: Any}:
        return {
            'time': packet.time,
            'x': packet.coordinates[0],
            'y': packet.coordinates[1],
            'z': packet.coordinates[2],
            'point': Point(*packet.coordinates),
            **{f'packet_{field}': value for field, value in packet.attributes.items()},
        }


class APRSDatabaseTable(PacketDatabaseTable, APRSPacketSource, APRSPacketSink):
    __aprs_fields = {
        'from': str,
        'to': str,
        'path': [str],
        'via': str,
        'timestamp': str,
        'symbol': str,
        'symbol_table': str,
        'latitude': float,
        'longitude': float,
        'altitude': float,
        'messagecapable': str,
        'format': str,
        'gpsfixstatus': str,
        'comment': str,
        'raw': str,
    }

    def __init__(
        self, hostname: str, database: str, table: str, callsigns: [str] = None, **kwargs
    ):
        """
        Create a new database connection to a table containing APRS packet data.

        :param hostname: hostname of database
        :param database: database name
        :param table: table name
        :param callsigns: list of callsigns to return from source
        """

        if 'fields' not in kwargs:
            kwargs['fields'] = self.__aprs_fields
        if 'primary_key' not in kwargs:
            kwargs['primary_key'] = ('time', 'packet_from')
        elif 'callsign' in kwargs['primary_key']:
            kwargs['primary_key'][kwargs['primary_key'].index('callsign')] = 'packet_from'
        kwargs['fields'] = {
            f'packet_{field}': field_type for field, field_type in kwargs['fields'].items()
        }
        PacketDatabaseTable.__init__(
            self, hostname=hostname, database=database, table=table, **kwargs
        )
        location = f'postgres://{self.hostname}:{self.port}/{self.database}/{self.name}'
        APRSPacketSource.__init__(self, location, callsigns)
        APRSPacketSink.__init__(self, location)
        self.__last_access_time = None

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        packets = []
        for record in self.records:
            record = {
                field
                if field in ['time', 'x', 'y', 'z']
                else field.replace('packet_', ''): value
                for field, value in record.items()
                if field not in ['point']
            }
            record['from_callsign'] = record['from']
            del record['from']
            if 'to' in record:
                record['to_callsign'] = record['to']
                del record['to']
            else:
                record['to'] = None
            if record['source'] is None:
                record['source'] = self.location
            packets.append(APRSPacket(**record))

        self.__last_access_time = datetime.now()
        return packets

    def __getitem__(self, key: (datetime, str)) -> APRSPacket:
        packet = PacketDatabaseTable.__getitem__(self, key)
        attributes = packet.attributes
        from_callsign = attributes['from']
        del attributes['from']
        if 'to' in attributes:
            to_callsign = attributes['to']
            del attributes['to']
        else:
            to_callsign = None
        return APRSPacket(
            from_callsign,
            to_callsign,
            packet.time,
            *packet.coordinates,
            packet.crs,
            **attributes,
        )

    def __setitem__(self, key: (datetime, str), packet: APRSPacket):
        PacketDatabaseTable.__setitem__(self, key, packet)

    def __contains__(self, packet: APRSPacket) -> bool:
        return PacketDatabaseTable.__contains__(self, packet)

    def send(self, packets: [APRSPacket]):
        PacketDatabaseTable.send(self, packets)

    def insert(self, packets: [APRSPacket]):
        PacketDatabaseTable.insert(self, packets)

    @property
    def records(self) -> [{str: Any}]:
        if self.callsigns is not None:
            return self.records_where({'packet_from': self.callsigns})
        else:
            return self.records_where(None)

    @staticmethod
    def __packet_record(packet: LocationPacket) -> {str: Any}:
        return {
            'time': packet.time,
            'callsign': packet.callsign,
            'x': packet.coordinates[0],
            'y': packet.coordinates[1],
            'z': packet.coordinates[2],
            'point': Point(*packet.coordinates),
            **{f'packet_{field}': value for field, value in packet.attributes.items()},
        }


class APRSis(APRSPacketSink, APRSPacketSource, NetworkConnection):
    def __init__(self, callsigns: [str] = None, hostname: str = None):
        if hostname is not None:
            self.__hostname, self.__port = split_hostname_port(hostname)
        else:
            self.__hostname, self.__port = ('rotate.aprs.net', 10152)

        NetworkConnection.__init__(self, f'{self.hostname}:{self.port}')
        APRSPacketSource.__init__(self, f'{self.hostname}:{self.port}', callsigns)

        if not self.connected:
            raise ConnectionError(f'no network connection')

        self.__send_buffer = []

    @property
    def hostname(self) -> str:
        return self.__hostname

    @property
    def port(self) -> int:
        return self.__port

    def send(self, packets: [APRSPacket]):
        if not isinstance(packets, Sequence) or isinstance(packets, str):
            packets = [packets]
        packets = [
            packet if not isinstance(packet, str) else APRSPacket.from_frame(packet)
            for packet in packets
        ]

        if len(self.__send_buffer) > 0:
            packets.extend(self.__send_buffer)
            self.__send_buffer.clear()

        callsigns = {packet.from_callsign for packet in packets}
        packets = {
            callsign: [packet for packet in packets if packet.from_callsign == callsign]
            for callsign in callsigns
        }

        if len(packets) > 0:
            LOGGER.info(f'sending {len(packets)} packet(s) to {self.location}: {packets}')
            for callsign, callsign_packets in packets.items():
                try:
                    frames = [packet.frame for packet in callsign_packets]
                    aprs_is = aprslib.IS(
                        callsign, aprslib.passcode(callsign), self.hostname, self.port
                    )
                    aprs_is.connect()
                    if len(frames) > 0:
                        aprs_is.sendall(r'\rn'.join(frames))
                    aprs_is.close()
                except ConnectionError as error:
                    LOGGER.info(
                        f'could not send packet(s) ({error}); reattempting on next iteration'
                    )
                    self.__send_buffer.extend(packets)

    @property
    def packets(self) -> [APRSPacket]:
        if self.callsigns is None:
            raise ConnectionError(
                f'cannot retrieve packets from APRS-IS with no callsigns specified'
            )

        packets = []

        def add_frames(frame: str):
            try:
                packet = APRSPacket.from_frame(frame)
                if packet.from_callsign in self.callsigns and packet not in packets:
                    packets.append(packet)
            except InvalidPacketError:
                pass

        aprs_is = aprslib.IS('NOCALL', '-1', self.hostname, self.port)

        aprs_is.connect()
        aprs_is.consumer(callback=add_frames, raw=True, blocking=False)
        aprs_is.close()

        return packets

    @property
    def connected(self) -> bool:
        aprs_is = aprslib.IS('NOCALL', '-1', self.hostname, self.port)
        for _ in range(5):
            try:
                aprs_is.connect()
                aprs_is.close()
                return True
            except:
                sleep(1)
                continue
        else:
            return False

    def close(self):
        pass



import numpy

# `dh/dt` based on historical flight data
FREEFALL_DESCENT_RATE = lambda altitude: -5.8e-08 * altitude ** 2 - 6.001
# TODO: propagate uncertainty
FREEFALL_DESCENT_RATE_UNCERTAINTY = lambda altitude: 0.2 * FREEFALL_DESCENT_RATE(altitude)

# integration of `(1/(dh/dt)) dh` based on historical flight data
# TODO make this model better with ML
FREEFALL_SECONDS_TO_GROUND = lambda altitude: 1695.02 * numpy.arctan(9.8311e-5 * altitude)



from argparse import ArgumentParser
from datetime import datetime, timedelta
from getpass import getpass
from logging import Logger
from pathlib import Path
import sys
import time

import humanize as humanize
import numpy
from tablecrow.utilities import convert_value, parse_hostname

from packetraven.connections import (
    APRSDatabaseTable,
    APRSfi,
    APRSis,
    PacketDatabaseTable,
    PacketGeoJSON,
    RawAPRSTextFile,
    SerialTNC,
    TimeIntervalError,
)
from packetraven.connections.base import PacketSource
from packetraven.packets import APRSPacket
from packetraven.packets.tracks import APRSTrack, LocationPacketTrack
from packetraven.packets.writer import write_packet_tracks
from packetraven.predicts import get_predictions, PredictionAPIURL, PredictionError
from packetraven.utilities import get_logger, read_configuration, repository_root

LOGGER = get_logger('packetraven', log_format='%(asctime)s | %(levelname)-8s | %(message)s')

CREDENTIALS_FILENAME = repository_root() / 'credentials.config'
DEFAULT_INTERVAL_SECONDS = 20


def main():
    args_parser = ArgumentParser()
    args_parser.add_argument('--callsigns', help='comma-separated list of callsigns to track')
    args_parser.add_argument(
        '--aprsfi-key', help='APRS.fi API key (from https://aprs.fi/page/api)'
    )
    args_parser.add_argument(
        '--tnc',
        help='comma-separated list of serial ports / text files of a TNC parsing APRS packets from analog audio to ASCII'
        ' (set to `auto` to use the first open serial port)',
    )
    args_parser.add_argument(
        '--database', help='PostGres database table `user@hostname:port/database/table`'
    )
    args_parser.add_argument('--tunnel', help='SSH tunnel `user@hostname:port`')
    args_parser.add_argument(
        '--igate', action='store_true', help='send new packets to APRS-IS'
    )
    args_parser.add_argument('--start', help='start date / time, in any common date format')
    args_parser.add_argument('--end', help='end date / time, in any common date format')
    args_parser.add_argument('--log', help='path to log file to save log messages')
    args_parser.add_argument('--output', help='path to output file to save packets')
    args_parser.add_argument(
        '--prediction-output',
        help='path to output file to save most up-to-date predicted trajectory',
    )
    args_parser.add_argument(
        '--prediction-start-location', help='start location to use for prediction (x,y,z)'
    )
    args_parser.add_argument(
        '--prediction-start-time', help='start time to use for prediction'
    )
    args_parser.add_argument(
        '--prediction-ascent-rate', help='ascent rate to use for prediction (m/s)'
    )
    args_parser.add_argument(
        '--prediction-burst-altitude', help='burst altitude to use for prediction (m)'
    )
    args_parser.add_argument(
        '--prediction-descent-rate', help='descent rate to use for prediction (m/s)'
    )
    args_parser.add_argument(
        '--prediction-float-altitude', help='float altitude to use for prediction (m)'
    )
    args_parser.add_argument('--prediction-float-duration', help='duration of float (s)')
    args_parser.add_argument(
        '--prediction-api',
        help=f'API URL to use for prediction (one of {[entry.value for entry in PredictionAPIURL]})',
    )
    args_parser.add_argument(
        '--interval',
        default=DEFAULT_INTERVAL_SECONDS,
        type=float,
        help=f'seconds between each main loop (default: {DEFAULT_INTERVAL_SECONDS})',
    )
    args_parser.add_argument(
        '--gui', action='store_true', help='start the graphical interface'
    )

    args = args_parser.parse_args()

    using_gui = args.gui
    using_igate = args.igate

    if args.callsigns is not None:
        callsigns = [callsign.upper() for callsign in args.callsigns.strip('"').split(',')]
    else:
        callsigns = None

    kwargs = {}

    if args.aprsfi_key is not None:
        kwargs['aprs_fi_key'] = args.aprsfi_key

    if args.tnc is not None:
        kwargs['tnc'] = [tnc.strip() for tnc in args.tnc.split(',')]

    if args.database is not None:
        database = parse_hostname(args.database)

        hostname = database['hostname']
        port = database['port']
        username = database['username']
        password = database['password']

        if port is not None:
            hostname = f'{hostname}:{port}'

        if hostname.count('/') != 2:
            LOGGER.error(
                f'unable to parse hostname, database name, and table name from input "{database}"'
            )
        else:
            (
                kwargs['database_hostname'],
                kwargs['database_database'],
                kwargs['database_table'],
            ) = hostname.split('/')
            kwargs['database_username'] = username
            kwargs['database_password'] = password

    if args.tunnel is not None:
        tunnel = parse_hostname(args.tunnel)

        hostname = tunnel['hostname']
        port = tunnel['port']
        username = tunnel['username']
        password = tunnel['password']

        if port is not None:
            hostname = f'{hostname}:{port}'

        kwargs['ssh_hostname'] = hostname
        kwargs['ssh_username'] = username
        kwargs['ssh_password'] = password

    start_date = (
        convert_value(args.start.strip('"'), datetime) if args.start is not None else None
    )
    end_date = convert_value(args.end.strip('"'), datetime) if args.end is not None else None

    if start_date is not None and end_date is not None:
        if start_date > end_date:
            temp_start_date = start_date
            start_date = end_date
            end_date = temp_start_date
            del temp_start_date

    if args.log is not None:
        log_filename = Path(args.log).expanduser()
        if log_filename.is_dir() or (not log_filename.exists() and log_filename.suffix == ''):
            log_filename = log_filename / f'packetraven_log_{datetime.now():%Y%m%dT%H%M%S}.txt'
        if not log_filename.parent.exists():
            log_filename.parent.mkdir(parents=True, exist_ok=True)
        get_logger(LOGGER.name, log_filename)
    else:
        log_filename = None

    if args.output is not None:
        output_filename = Path(args.output).expanduser()
        if output_filename.is_dir() or (
            not output_filename.exists() and output_filename.suffix == ''
        ):
            output_filename = (
                output_filename / f'packetraven_output_{datetime.now():%Y%m%dT%H%M%S}.geojson'
            )
        if not output_filename.parent.exists():
            output_filename.parent.mkdir(parents=True, exist_ok=True)
    else:
        output_filename = None

    if args.prediction_output is not None:
        prediction_filename = Path(args.prediction_output).expanduser()
        if prediction_filename.is_dir() or (
            not prediction_filename.exists() and prediction_filename.suffix == ''
        ):
            prediction_filename = (
                prediction_filename
                / f'packetraven_predict_{datetime.now():%Y%m%dT%H%M%S}.geojson'
            )
        if not prediction_filename.parent.exists():
            prediction_filename.parent.mkdir(parents=True, exist_ok=True)

        if args.prediction_start_location is not None:
            kwargs['prediction_start_location'] = [
                float(value) for value in args.prediction_start_location.split(',')
            ]

        if args.prediction_start_time is not None:
            kwargs['prediction_start_time'] = convert_value(
                args.prediction_start_time, datetime
            )

        if args.prediction_ascent_rate is not None:
            kwargs['prediction_ascent_rate'] = float(args.prediction_ascent_rate)

        if args.prediction_burst_altitude is not None:
            kwargs['prediction_burst_altitude'] = float(args.prediction_burst_altitude)

        if args.prediction_descent_rate is not None:
            kwargs['prediction_sea_level_descent_rate'] = float(args.prediction_descent_rate)

        if args.prediction_float_altitude is not None:
            kwargs['prediction_float_altitude'] = float(args.prediction_float_altitude)

        if args.prediction_float_duration is not None:
            kwargs['prediction_float_duration'] = convert_value(
                args.prediction_float_duration, timedelta
            )

        if args.prediction_api is not None:
            kwargs['prediction_api_url'] = args.prediction_api
    else:
        prediction_filename = None

    interval_seconds = convert_value(args.interval, timedelta) / timedelta(seconds=1)
    if interval_seconds < 1:
        interval_seconds = 1

    if CREDENTIALS_FILENAME.exists():
        credentials = kwargs.copy()
        for section in read_configuration(CREDENTIALS_FILENAME).values():
            credentials.update(section)
        kwargs = {
            key: value if key not in kwargs or kwargs[key] is None else kwargs[key]
            for key, value in credentials.items()
        }

    if using_gui:
        from packetraven.gui import PacketRavenGUI

        PacketRavenGUI(
            callsigns,
            start_date,
            end_date,
            log_filename,
            output_filename,
            prediction_filename,
            interval_seconds,
            using_igate,
            **kwargs,
        )
    else:
        connections = []
        if 'tnc' in kwargs:
            for tnc_location in kwargs['tnc']:
                tnc_location = tnc_location.strip()
                try:
                    if Path(tnc_location).suffix in ['.txt', '.log']:
                        tnc_location = RawAPRSTextFile(tnc_location, callsigns)
                        LOGGER.info(f'reading file {tnc_location.location}')
                        connections.append(tnc_location)
                    else:
                        tnc_location = SerialTNC(tnc_location, callsigns)
                        LOGGER.info(f'opened port {tnc_location.location}')
                        connections.append(tnc_location)
                except ConnectionError as error:
                    LOGGER.warning(f'{error.__class__.__name__} - {error}')

        if 'aprs_fi_key' in kwargs:
            aprs_fi_kwargs = {key: kwargs[key] for key in ['aprs_fi_key'] if key in kwargs}
            try:
                aprs_api = APRSfi(callsigns=callsigns, api_key=aprs_fi_kwargs['aprs_fi_key'])
                LOGGER.info(f'connected to {aprs_api.location}')
                connections.append(aprs_api)
            except ConnectionError as error:
                LOGGER.warning(f'{error.__class__.__name__} - {error}')

        if 'database_hostname' in kwargs:
            database_kwargs = {
                key: kwargs[key]
                for key in [
                    'database_hostname',
                    'database_database',
                    'database_table',
                    'database_username',
                    'database_password',
                ]
                if key in kwargs
            }
            ssh_tunnel_kwargs = {
                key: kwargs[key]
                for key in ['ssh_hostname', 'ssh_username', 'ssh_password']
                if key in kwargs
            }

            try:
                if 'ssh_hostname' in ssh_tunnel_kwargs:
                    if (
                        'ssh_username' not in ssh_tunnel_kwargs
                        or ssh_tunnel_kwargs['ssh_username'] is None
                    ):
                        ssh_username = input(
                            f'enter username for SSH host "{ssh_tunnel_kwargs["ssh_hostname"]}": '
                        )
                        if ssh_username is None or len(ssh_username) == 0:
                            raise ConnectionError('missing SSH username')
                        ssh_tunnel_kwargs['ssh_username'] = ssh_username

                    if (
                        'ssh_password' not in ssh_tunnel_kwargs
                        or ssh_tunnel_kwargs['ssh_password'] is None
                    ):
                        ssh_password = getpass(
                            f'enter password for SSH user "{ssh_tunnel_kwargs["ssh_username"]}": '
                        )
                        if ssh_password is None or len(ssh_password) == 0:
                            raise ConnectionError('missing SSH password')
                        ssh_tunnel_kwargs['ssh_password'] = ssh_password

                if (
                    'database_username' not in database_kwargs
                    or database_kwargs['database_username'] is None
                ):
                    database_username = input(
                        f'enter username for database '
                        f'"{database_kwargs["database_hostname"]}/{database_kwargs["database_database"]}": '
                    )
                    if database_username is None or len(database_username) == 0:
                        raise ConnectionError('missing database username')
                    database_kwargs['database_username'] = database_username

                if (
                    'database_password' not in database_kwargs
                    or database_kwargs['database_password'] is None
                ):
                    database_password = getpass(
                        f'enter password for database user "{database_kwargs["database_username"]}": '
                    )
                    if database_password is None or len(database_password) == 0:
                        raise ConnectionError('missing database password')
                    database_kwargs['database_password'] = database_password

                database = APRSDatabaseTable(
                    **{
                        key.replace('database_', ''): value
                        for key, value in database_kwargs.items()
                    },
                    **ssh_tunnel_kwargs,
                    callsigns=callsigns,
                )
                LOGGER.info(f'connected to {database.location}')
                connections.append(database)
            except ConnectionError:
                database = None
        else:
            database = None

        if len(connections) == 0:
            if output_filename is not None and output_filename.exists():
                connections.append(PacketGeoJSON(output_filename))
            else:
                LOGGER.error(f'no connections started')
                sys.exit(1)

        if using_igate:
            try:
                aprs_is = APRSis(callsigns)
            except ConnectionError:
                aprs_is = None
        else:
            aprs_is = None

        filter_message = 'retrieving packets'
        if start_date is not None and end_date is None:
            filter_message += f' sent after {start_date:%Y-%m-%d %H:%M:%S}'
        elif start_date is None and end_date is not None:
            filter_message += f' sent before {end_date:%Y-%m-%d %H:%M:%S}'
        elif start_date is not None and end_date is not None:
            filter_message += f' sent between {start_date:%Y-%m-%d %H:%M:%S} and {end_date:%Y-%m-%d %H:%M:%S}'
        if callsigns is not None:
            filter_message += f' from {len(callsigns)} callsigns: {callsigns}'
        LOGGER.info(filter_message)

        LOGGER.info(
            f'listening for packets every {interval_seconds}s from {len(connections)} connection(s): '
            f'{", ".join([connection.location for connection in connections])}'
        )

        packet_tracks = {}
        try:
            while len(connections) > 0:
                try:
                    new_packets = retrieve_packets(
                        connections,
                        packet_tracks,
                        database,
                        start_date=start_date,
                        end_date=end_date,
                        logger=LOGGER,
                    )

                    output_filename_index = None
                    for index, connection in enumerate(connections):
                        if isinstance(connection, PacketGeoJSON):
                            output_filename_index = index
                    if output_filename_index is not None:
                        connections.pop(output_filename_index)

                    if prediction_filename is not None:
                        try:
                            predictions = get_predictions(
                                packet_tracks,
                                **{
                                    key.replace('prediction_', ''): value
                                    for key, value in kwargs.items()
                                    if 'prediction_' in key
                                },
                            )
                            write_packet_tracks(predictions.values(), prediction_filename)
                        except PredictionError as error:
                            LOGGER.warning(f'{error.__class__.__name__} - {error}')
                        except Exception as error:
                            LOGGER.warning(
                                f'error retrieving prediction trajectory - {error.__class__.__name__} - {error}'
                            )

                except Exception as error:
                    LOGGER.exception(f'{error.__class__.__name__} - {error}')
                    new_packets = {}

                if len(new_packets) > 0:
                    if output_filename is not None:
                        write_packet_tracks(
                            [packet_tracks[callsign] for callsign in packet_tracks],
                            output_filename,
                        )
                    if aprs_is is not None:
                        for packets in new_packets.values():
                            aprs_is.send(packets)
                time.sleep(interval_seconds)
        except KeyboardInterrupt:
            for connection in connections:
                connection.close()
            sys.exit(0)


def retrieve_packets(
    connections: [PacketSource],
    packet_tracks: [LocationPacketTrack],
    database: PacketDatabaseTable = None,
    start_date: datetime = None,
    end_date: datetime = None,
    logger: Logger = None,
) -> {str: APRSPacket}:
    if logger is None:
        logger = LOGGER

    logger.debug(f'receiving packets from {len(connections)} source(s)')
    current_time = datetime.now()

    parsed_packets = []
    for connection in connections:
        try:
            connection_packets = connection.packets
            parsed_packets.extend(connection_packets)
        except ConnectionError as error:
            LOGGER.error(f'{connection.__class__.__name__} - {error}')
        except TimeIntervalError:
            pass

    logger.debug(f'received {len(parsed_packets)} packets')

    new_packets = {}
    if len(parsed_packets) > 0:
        updated_callsigns = set()
        for parsed_packet in parsed_packets:
            callsign = parsed_packet['callsign']

            if start_date is not None and parsed_packet.time <= start_date:
                continue
            if end_date is not None and parsed_packet.time >= end_date:
                continue

            if callsign not in packet_tracks:
                packet_tracks[callsign] = APRSTrack(callsign, [parsed_packet])
                logger.debug(f'started tracking callsign {callsign:8}')
            else:
                packet_track = packet_tracks[callsign]
                if parsed_packet not in packet_track:
                    packet_track.append(parsed_packet)
                else:
                    if database is None or parsed_packet.source != database.location:
                        logger.debug(f'skipping duplicate packet: {parsed_packet}')
                    continue

            if parsed_packet.source not in new_packets:
                new_packets[parsed_packet.source] = []
            new_packets[parsed_packet.source].append(parsed_packet)
            if callsign not in updated_callsigns:
                updated_callsigns.add(callsign)

        for source in new_packets:
            new_packets[source] = list(sorted(new_packets[source]))

        for source, packets in new_packets.items():
            logger.info(f'received {len(packets)} new packet(s) from {source}')

        if database is not None:
            for packets in new_packets.values():
                database.send(
                    packet for packet in packets if packet.source != database.location
                )

        updated_callsigns = sorted(updated_callsigns)
        for callsign in updated_callsigns:
            packet_track = packet_tracks[callsign]
            packet_time = datetime.utcfromtimestamp(
                (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00Z'))
                / numpy.timedelta64(1, 's')
            )
            packet_track.sort()
            try:
                coordinate_string = ', '.join(
                    f'{coordinate:.3f}°' for coordinate in packet_track.coordinates[-1, :2]
                )
                logger.info(
                    f'{callsign:9} - packet #{len(packet_track):<3} - ({coordinate_string}, {packet_track.coordinates[-1, 2]:9.2f}m)'
                    f'; packet time is {packet_time} ({humanize.naturaltime(current_time - packet_time)}, {packet_track.intervals[-1]:6.1f} s interval)'
                    f'; traveled {packet_track.overground_distances[-1]:6.1f} m ({packet_track.ground_speeds[-1]:5.1f} m/s) over the ground'
                    f', and {packet_track.ascents[-1]:6.1f} m ({packet_track.ascent_rates[-1]:5.1f} m/s) vertically, since the previous packet'
                )
            except Exception as error:
                logger.exception(f'{error.__class__.__name__} - {error}')

        for callsign in updated_callsigns:
            packet_track = packet_tracks[callsign]
            packet_time = datetime.utcfromtimestamp(
                (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00Z'))
                / numpy.timedelta64(1, 's')
            )
            try:
                message = (
                    f'{callsign:9} - '
                    f'altitude: {packet_track.altitudes[-1]:6.1f} m'
                    f'; avg. ascent rate: {numpy.mean(packet_track.ascent_rates[packet_track.ascent_rates > 0]):5.1f} m/s'
                    f'; avg. descent rate: {numpy.mean(packet_track.ascent_rates[packet_track.ascent_rates < 0]):5.1f} m/s'
                    f'; avg. ground speed: {numpy.mean(packet_track.ground_speeds):5.1f} m/s'
                    f'; avg. packet interval: {numpy.mean(packet_track.intervals):6.1f} s'
                )

                if packet_track.time_to_ground >= timedelta(seconds=0):
                    landing_time = packet_time + packet_track.time_to_ground
                    time_to_ground = current_time - landing_time
                    message += (
                        f'; estimated landing: {landing_time:%Y-%m-%d %H:%M:%S} ({humanize.naturaltime(time_to_ground)})'
                        f'; max altitude: {packet_track.coordinates[:, 2].max():.2f} m'
                    )

                logger.info(message)

            except Exception as error:
                logger.exception(f'{error.__class__.__name__} - {error}')

    return new_packets


if __name__ == '__main__':
    main()



from datetime import datetime
from os import PathLike
from pathlib import Path
from urllib.parse import urlparse

from dateutil.parser import parse as parse_date
import geojson
import requests

from packetraven.connections.base import (
    APRSPacketSource,
    LOGGER,
    PacketSource,
    TimeIntervalError,
)
from packetraven.packets import APRSPacket, LocationPacket


class RawAPRSTextFile(APRSPacketSource):
    def __init__(self, filename: PathLike = None, callsigns: str = None):
        """
        read APRS packets from a given text file where each line consists of the time sent (`YYYY-MM-DDTHH:MM:SS`) followed by
        a colon `:` and then the raw APRS string

        :param filename: path to text file
        :param callsigns: list of callsigns to return from source
        """

        if not urlparse(str(filename)).scheme in ['http', 'https', 'ftp', 'sftp']:
            if not isinstance(filename, Path):
                if isinstance(filename, str):
                    filename = filename.strip('"')
                filename = Path(filename)
            filename = str(filename)

        super().__init__(filename, callsigns)
        self.__last_access_time = None
        self.__parsed_lines = []

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        if Path(self.location).exists():
            file_connection = open(Path(self.location).expanduser().resolve())
            lines = file_connection.readlines()
        else:
            file_connection = requests.get(self.location, stream=True)
            lines = file_connection.iter_lines()

        packets = []
        for line in lines:
            if len(line) > 0:
                if isinstance(line, bytes):
                    line = line.decode()
                if line not in self.__parsed_lines:
                    self.__parsed_lines.append(line)
                    try:
                        packet_time, raw_aprs = line.split(': ', 1)
                        packet_time = parse_date(packet_time)
                    except:
                        raw_aprs = line
                        packet_time = datetime.now()
                    raw_aprs = raw_aprs.strip()
                    try:
                        packets.append(
                            APRSPacket.from_frame(raw_aprs, packet_time, source=self.location)
                        )
                    except Exception as error:
                        LOGGER.error(f'{error.__class__.__name__} - {error}')

        file_connection.close()

        if self.callsigns is not None:
            packets = [packet for packet in packets if packet.from_callsign in self.callsigns]
        self.__last_access_time = datetime.now()

        return packets

    def close(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.location)}, {repr(self.callsigns)})'


class PacketGeoJSON(PacketSource):
    def __init__(self, filename: PathLike = None):
        """
        read location packets from a given GeoJSON file

        :param filename: path to GeoJSON file
        """

        if not urlparse(str(filename)).scheme in ['http', 'https', 'ftp', 'sftp']:
            if not isinstance(filename, Path):
                if isinstance(filename, str):
                    filename = filename.strip('"')
                filename = Path(filename)
            filename = str(filename)

        super().__init__(filename)
        self.__last_access_time = None

    @property
    def packets(self) -> [LocationPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        if Path(self.location).exists():
            with open(Path(self.location).expanduser().resolve()) as file_connection:
                features = geojson.load(file_connection)
        else:
            response = requests.get(self.location, stream=True)
            features = geojson.loads(response.text)

        packets = []
        for feature in features['features']:
            if feature['geometry']['type'] == 'Point':
                properties = feature['properties']
                time = parse_date(properties['time'])
                del properties['time']

                if 'from' in properties:
                    from_callsign = properties['from']
                    to_callsign = properties['to']
                    del properties['from'], properties['to']

                    packet = APRSPacket(
                        from_callsign,
                        to_callsign,
                        time,
                        *feature['geometry']['coordinates'],
                        source=self.location,
                        **properties,
                    )
                else:
                    packet = LocationPacket(
                        time,
                        *feature['geometry']['coordinates'],
                        source=self.location,
                        **properties,
                    )

                packets.append(packet)

        self.__last_access_time = datetime.now()

        return packets

    def close(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.location)})'



#!/usr/bin/env python
import importlib
import logging
import os
from pathlib import Path
import re
import subprocess
import sys

from setuptools import config, find_packages, setup

DEPENDENCIES = {
    'aprslib': [],
    'humanize': [],
    'numpy': [],
    'pandas': ['numpy'],
    'pyserial': [],
    'geojson': [],
    'fastkml': [],
    'matplotlib': [],
    'psycopg2-binary': [],
    'pyproj': [],
    'requests': [],
    'shapely': [],
    'sshtunnel': [],
    'tablecrow>=1.3.9': [],
    'teek': [],
}

installed_packages = [
    re.split('#egg=', re.split('==| @ ', package.decode())[0])[-1].lower()
    for package in subprocess.check_output(
        [sys.executable, '-m', 'pip', 'freeze']
    ).splitlines()
]

missing_dependencies = {
    dependency: subdependencies
    for dependency, subdependencies in DEPENDENCIES.items()
    if re.split('<|<=|==|>=|>', dependency)[0].lower() not in installed_packages
}

if (Path(sys.prefix) / 'conda-meta').exists():
    for dependency in list(missing_dependencies):
        try:
            subprocess.check_call(['conda', 'install', '-y', dependency])
            del missing_dependencies[dependency]
        except:
            continue

if os.name == 'nt':
    if len(missing_dependencies) > 0:
        try:
            import pipwin
        except:
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'pipwin'])
            subprocess.check_call([sys.executable, '-m', 'pipwin', 'refresh'])

    for dependency, subdependencies in list(missing_dependencies.items()):
        failed_pipwin_packages = []
        for _ in range(1 + len(subdependencies)):
            for package_name in [dependency] + subdependencies:
                try:
                    importlib.import_module(package_name)
                except:
                    try:
                        subprocess.check_call(
                            [sys.executable, '-m', 'pipwin', 'install', package_name.lower()]
                        )
                        if package_name in failed_pipwin_packages:
                            failed_pipwin_packages.remove(package_name)
                        if package_name in missing_dependencies:
                            del missing_dependencies[package_name]
                    except subprocess.CalledProcessError:
                        failed_pipwin_packages.append(package_name)

            # since we don't know the dependencies here, repeat this process n number of times
            # (worst case is `O(n)`, where the first package is dependant on all the others)
            if len(failed_pipwin_packages) == 0:
                break

try:
    try:
        from dunamai import Version
    except ImportError:
        import subprocess
        import sys

        subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'dunamai'])
        from dunamai import Version

    version = Version.from_any_vcs().serialize()
except RuntimeError as error:
    logging.exception(error)
    version = '0.0.0'

metadata = config.read_configuration('setup.cfg')['metadata']

setup(
    name=metadata['name'],
    version=version,
    author=metadata['author'],
    author_email=metadata['author_email'],
    description=metadata['description'],
    long_description=metadata['long_description'],
    long_description_content_type='text/markdown',
    url=metadata['url'],
    packages=find_packages(),
    python_requires='>=3.8',
    setup_requires=['dunamai', 'setuptools>=41.2'],
    install_requires=list(DEPENDENCIES),
    extras_require={
        'testing': ['pytest', 'pytest-cov', 'pytest-xdist', 'pytz'],
        'development': ['flake8', 'isort', 'oitnb', 'wheel'],
    },
    entry_points={'console_scripts': ['packetraven=packetraven.__main__:main']},
)



from packetraven.connections.base import TimeIntervalError
from packetraven.connections.file import PacketGeoJSON, RawAPRSTextFile
from packetraven.connections.internet import (
    APRSDatabaseTable,
    APRSfi,
    APRSis,
    PacketDatabaseTable,
)
from packetraven.connections.serial import SerialTNC
